# -*- coding: utf-8 -*-
"""
Created on Tue Apr 21 12:05:22 2015

@author:    charleslberube@gmail.com
            École Polytechnique de Montréal

The MIT License (MIT)

Copyright (c) 2016 Charles L. Bérubé

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.

https://opensource.org/licenses/MIT
https://github.com/clberube/bisip

This python module may be used to import SIP data, run MCMC inversion and
return the results.

It is imported as:

Call with minimal arguments:

sol = mcmcinv('ColeCole', '/Documents/DataFiles/DATA.dat')

Call with all optional arguments:

sol = mcmcinv( model='ColeCole', filename='/Documents/DataFiles/DATA.dat',
                 mcmc=mcmc_params, headers=1, ph_units='mrad', cc_modes=2,
                 debye_poly=4, c_exp = 1.0, keep_traces=False)
"""
from __future__ import division
from __future__ import print_function

#==============================================================================
# Import PyMC, Numpy, and Cython extension
from builtins import str
from builtins import range
from past.utils import old_div
import pymc
import numpy as np
from bisip.cython_funcs import ColeCole_cyth1, ColeCole_cyth2, Dias_cyth, Decomp_cyth, Shin_cyth
# Imports to save things
from os import path, makedirs
from os import getcwd
from sys import argv
from datetime import datetime

#==============================================================================
# Function to run MCMC simulation on selected model
# Arguments: model <function>, mcmc parameters <dict>,traces path <string>
def run_MCMC(function, mc_p, save_traces=False, save_where=None):
    print("\nMCMC parameters:\n", mc_p)
    if save_traces:
        # If path doesn't exist, create it
        if not path.exists(save_where): makedirs(save_where)
        MDL = pymc.MCMC(function, db='txt',
                        dbname=save_where)
    else:
        MDL = pymc.MCMC(function, db='ram',
                        dbname=save_where)

    if mc_p["adaptive"]:
        if mc_p['verbose']:
            mc_p['verbose'] = 1
        MDL.use_step_method(pymc.AdaptiveMetropolis, MDL.stochastics, delay=mc_p["cov_delay"], interval=mc_p['cov_inter'], shrink_if_necessary=True, verbose=mc_p['verbose'])

    else:
        for stoc in MDL.stochastics:
            MDL.use_step_method(pymc.Metropolis, stoc,
                                proposal_distribution='Normal',
                                scale=mc_p['prop_scale'], verbose=mc_p['verbose'])

    for i in range(1, mc_p['nb_chain']+1):
        print('\n Chain #%d/%d'%(i, mc_p['nb_chain']))
        MDL.sample(mc_p['nb_iter'], mc_p['nb_burn'], mc_p['thin'], tune_interval=mc_p['tune_inter'], tune_throughout=False)
    return MDL

#==============================================================================
# To import data
# Arguments: file name, number of header lines to skip, phase units
def get_data(filename,headers,ph_units):
    # Importation des données .DAT
    dat_file = np.loadtxt("%s"%(filename),skiprows=headers,delimiter=',')
    labels = ["freq", "amp", "pha", "amp_err", "pha_err"]
    data = {l:dat_file[:,i] for (i,l) in enumerate(labels)}
    if ph_units == "mrad":
        data["pha"] = old_div(data["pha"],1000)                      # mrad to rad
        data["pha_err"] = old_div(data["pha_err"],1000)              # mrad to rad
    if ph_units == "deg":
        data["pha"] = np.radians(data["pha"])               # deg to rad
        data["pha_err"] = np.radians(data["pha_err"])       # deg to rad
    data["phase_range"] = abs(max(data["pha"])-min(data["pha"])) # Range of phase measurements (used in NRMS error calculation)
    data["Z"]  = data["amp"]*(np.cos(data["pha"]) + 1j*np.sin(data["pha"]))
    EI = np.sqrt(((data["amp"]*np.cos(data["pha"])*data["pha_err"])**2)+(np.sin(data["pha"])*data["amp_err"])**2)
    ER = np.sqrt(((data["amp"]*np.sin(data["pha"])*data["pha_err"])**2)+(np.cos(data["pha"])*data["amp_err"])**2)
    data["Z_err"] = ER + 1j*EI
    # Normalization of amplitude
    data["Z_max"] = max(abs(data["Z"]))  # Maximum amplitude
    zn, zn_e = old_div(data["Z"],data["Z_max"]), old_div(data["Z_err"],data["Z_max"]) # Normalization of impedance by max amplitude
    data["zn"] = np.array([zn.real, zn.imag]) # 2D array with first column = real values, second column = imag values
    data["zn_err"] = np.array([zn_e.real, zn_e.imag]) # 2D array with first column = real values, second column = imag values
    return data

#==============================================================================
# To extract important information from the model (MDL)
# Used at the end of inversion routine
# Arguments: model <pymc model object>, maximum amplitude measured <float>
def format_results(M, Z_max):
    var_keys = [s.__name__ for s in M.stochastics] + [d.__name__ for d in M.deterministics]
    var_keys = [s for s in var_keys if s not in ["zmod", "mp"]]
    Mst = M.stats(chain=-1)
    pm = {k: Mst[k]["mean"] for k in var_keys}
    pm.update({k+"_std": Mst[k]["standard deviation"] for k in var_keys})
    pm.update({"R0": Z_max*pm["R0"],"R0_std": Z_max*pm["R0_std"]}) # remove normalization
    pm.update({k.replace("log_", ""): 10**pm[k] for k in var_keys if k.startswith("log_")})
    pm.update({(k.replace("log_", ""))+"_std": abs(pm[k+"_std"]/pm[k])*(10**pm[k]) for k in var_keys if k.startswith("log_")})
    pm = {k: v for (k, v) in list(pm.items()) if "log_" not in k}
    return pm           # returns parameters and uncertainty

#==============================================================================
# Main inversion function.
# Import in script using
# from bisip import mcmcinv
# Default MCMC parameters:
mcmc_params = {"adaptive"   : False,
               "nb_chain"   : 1,
               "nb_iter"    : 100000,
               "nb_burn"    : 80000,
               "thin"       : 1,
               "tune_inter" : 10000,
               "prop_scale" : 1.0,
               "verbose"    : False,
               "cov_inter"  : 10000,
               "cov_delay"  : 10000,
                }

def mcmcinv(model, filename, mcmc=mcmc_params, headers=1,
               ph_units="mrad", cc_modes=2, decomp_poly=4, c_exp=1.0, log_min_tau=-3, keep_traces=False):

#==============================================================================
    """Cole-Cole Bayesian Model"""
#==============================================================================
    def ColeColeModel(cc_modes):
        # Initial guesses
        p0 = {'R0'       : 1.0,
              'm'        : None,
              'log_tau'  : None,
              'c'        : None,
              }
        # Stochastics
        R0 = pymc.Uniform('R0', lower=0.7, upper=1.3 , value=p0["R0"])
        m = pymc.Uniform('m', lower=0.0, upper=1.0, value=p0["m"], size=cc_modes)
        log_tau = pymc.Uniform('log_tau', lower=-7.0, upper=4.0, value=p0['log_tau'], size=cc_modes)
        c = pymc.Uniform('c', lower=0.0, upper=1.0, value=p0['c'], size=cc_modes)
        # Deterministics
        @pymc.deterministic(plot=False)
        def zmod(cc_modes=cc_modes, R0=R0, m=m, lt=log_tau, c=c):
            return ColeCole_cyth1(w, R0, m, lt, c)
        # Likelihood
        obs = pymc.Normal('obs', mu=zmod, tau=old_div(1.0,(data["zn_err"]**2)), value=data["zn"], size=(2,len(w)), observed=True)
        return locals()

#==============================================================================
    """Shin Bayesian Model"""
#==============================================================================
    def ShinModel():
        # Initial guesses
        p0 = {'R'      : [0.5, 0.5],
              'log_Q'  : [0,-4],
              'n'      : [0.5, 0.5],
              'log_tau': None,
              'm'      : None,
              }
        # Stochastics
        R = pymc.Uniform('R', lower=0.0, upper=1.0, value=p0["R"], size=2)
        log_Q = pymc.Uniform('log_Q', lower=-7, upper=2, value=p0["log_Q"], size=2)
        n = pymc.Uniform('n', lower=0.0, upper=1.0, value=p0["n"], size=2)
        # Deterministics
        @pymc.deterministic(plot=False)
        def zmod(R=R, log_Q=log_Q, n=n):
            return Shin_cyth(w, R, log_Q, n)
        @pymc.deterministic(plot=False)
        def log_tau(R=R, log_Q=log_Q, n=n):
            return np.log10((R*(10**log_Q))**(old_div(1.,n)))
        @pymc.deterministic(plot=False)
        def R0(R=R):
            return R[0]+R[1]
        @pymc.deterministic(plot=False)
        def m(R=R):
            return seigle_m*( old_div(max(R), (max(R) + min(R))))
        #Likelihood
        obs = pymc.Normal('obs', mu=zmod, tau=old_div(1.0,(data["zn_err"]**2)), value=data["zn"], size = (2,len(w)), observed=True)
        return locals()

#==============================================================================
    """Dias Bayesian Model"""
#==============================================================================
    def DiasModel():
        # Initial guesses
        p0 = {'R0'     :  1.0,
              'm'      :  seigle_m,
              'log_tau':  None,
              'eta'    :  None,
              'delta'  :  None,
              }
        # Stochastics
        R0 = pymc.Uniform('R0', lower=0.9, upper=1.1 , value=1)
        m = pymc.Uniform('m', lower=0.0, upper=1.0, value=p0['m'])
        log_tau = pymc.Uniform('log_tau', lower=-7.0, upper=0.0, value=p0['log_tau'])
        eta = pymc.Uniform('eta', lower=0.0, upper=50.0, value=p0['eta'])
        delta = pymc.Uniform('delta', lower=0.0, upper=1.0, value=p0['delta'])
        # Deterministics
        @pymc.deterministic(plot=False)
        def zmod(R0=R0, m=m, lt=log_tau, eta=eta, delta=delta):
            return Dias_cyth(w, R0, m, lt, eta, delta)
        # Likelihood
        obs = pymc.Normal('obs', mu=zmod, tau=old_div(1.0,(data["zn_err"]**2)), value=data["zn"], size = (2,len(w)), observed=True)
        return locals()

#==============================================================================
    """Debye, Warburg, Cole-Cole decomposition Bayesian Model"""
#==============================================================================
    def PolyDecompModel(decomp_poly, c_exp):
        # Initial guesses
        p0 = {'R0'         : 1.0,
              'a'          : None,
#              'a'          : ([0.01, -0.01, -0.01, 0.001, 0.001]+[0.0]*(decomp_poly-4))[:(decomp_poly+1)],
              'log_tau_hi' : -5.0,
              'm_hi'       : 0.5,
              'TotalM'     : None,
              'log_MeanTau': None,
              'U'          : None,
              }
        # Stochastics
        R0 = pymc.Uniform('R0', lower=0.7, upper=1.3, value=p0['R0'])
#        m_hi = pymc.Uniform('m_hi', lower=0.0, upper=1.0, value=p0['m_hi'])
#        log_tau_hi = pymc.Uniform('log_tau_hi', lower=-8.0, upper=-3.0, value=p0['log_tau_hi'])
#        a = pymc.Uniform('a', lower=-0.1, upper=0.1, value=p0["a"], size=decomp_poly+1)
        a = pymc.Normal('a', mu=0, tau=old_div(1.,(0.01**2)), value=p0["a"], size=decomp_poly+1)

        # Deterministics
#        @pymc.deterministic(plot=False)
#        def m_hi(mp_hi=mp_hi):
#            return 10**mp_hi / (1 + 10**mp_hi)
        @pymc.deterministic(plot=False)
        def zmod(R0=R0, a=a):
            return Decomp_cyth(w, tau_10, log_taus, c_exp, R0, a)
        @pymc.deterministic(plot=False)
        def m_(a=a):
            return np.sum((a*log_taus.T).T, axis=0)
        @pymc.deterministic(plot=False)
        def total_m(m_=m_):
            return np.sum(m_[(log_tau >= log_min_tau)&(m_ >= 0)&(log_tau <= max(log_tau)-1)])
        @pymc.deterministic(plot=False)
        def log_half_tau(m_=m_):
            return log_tau[cond][np.where(np.cumsum(m_[cond])/np.sum(m_[cond]) > 0.5)[0][0]]
        @pymc.deterministic(plot=False)
        def log_peak_tau(m_=m_):
            cond = np.r_[True, m_[1:] > m_[:-1]] & np.r_[m_[:-1] > m_[1:], True]
            cond[0] = False
            try: return log_tau[cond][0]
            except: return log_tau[0]
        @pymc.deterministic(plot=False)
        def log_mean_tau(m_=m_):
            return np.log10(np.exp(old_div(np.sum(m_[cond]*np.log(10**log_tau[cond])),np.sum(m_[cond]))))
        # Likelihood
        obs = pymc.Normal('obs', mu=zmod, tau=old_div(1.0,(data["zn_err"]**2)), value=data["zn"], size = (2, len(w)), observed=True)
#        obs = pymc.Normal('obs', mu=zmod[1], tau=1.0/(data["zn_err"][1]**2), value=data["zn"][1], size = len(w), observed=True)
        return locals()

#==============================================================================
    """
    Main section
    """
#==============================================================================
    # Importing data
    data = get_data(filename, headers, ph_units)
    seigle_m = (old_div((data["amp"][-1] - data["amp"][0]), data["amp"][-1]) ) # Estimating Seigel chargeability
    w = 2*np.pi*data["freq"] # Frequencies measured in rad/s
#    n_freq = len(w)
#    n_decades = np.ceil(max(np.log10(old_div(1.0,w)))) - np.floor(min(np.log10(old_div(1.0,w))))
    # Relaxation times associated with the measured frequencies (Debye decomposition only)
    log_tau = np.linspace(np.floor(min(np.log10(old_div(1.0,w)))-1), np.floor(max(np.log10(old_div(1.0,w)))+1), 50)
    cond = (log_tau >= min(log_tau)+1)&(log_tau <= max(log_tau)-1)
    log_taus = np.array([log_tau**i for i in range(0,decomp_poly+1,1)]) # Polynomial approximation for the RTD
    tau_10 = 10**log_tau # Accelerates sampling
    data["tau"] = tau_10 # Put relaxation times in data dictionary

    # Time and date (for saving traces)
    sample_name = filename.replace("\\", "/").split("/")[-1].split(".")[0]
#    actual_path = str(path.dirname(path.realpath(argv[0])))
    working_path = getcwd().replace("\\", "/")+"/"
    now = datetime.now()
    save_time = now.strftime('%Y%m%d_%H%M%S')
    save_date = now.strftime('%Y%m%d')
    out_path = '%s/Txt traces/%s/%s/%s-%s-%s/'%(working_path, save_date,
                                                 sample_name, model,
                                                 sample_name, save_time)

    """
    #==========================================================================
    Call to run_MCMC function
    #==========================================================================
    """
    # "ColeCole", "Dias", "Debye" or "Shin"
    sim_dict = {"ColeCole": {"func": ColeColeModel,     "args": [cc_modes]          },
                "Dias":     {"func": DiasModel,         "args": []                  },
                "PDecomp":  {"func": PolyDecompModel,   "args": [decomp_poly, c_exp]},
                "Shin":     {"func": ShinModel,         "args": []                  },
#                "Custom":   {"func": YourModel,     "args": [opt_args]   },
                }
    simulation = sim_dict[model] # Pick entries for the selected model
    MDL = run_MCMC(simulation["func"](*simulation["args"]), mcmc, save_traces=keep_traces, save_where=out_path) # Run MCMC simulation with selected model and arguments
#    if not keep_traces: rmtree(out_path)   # Deletes the traces if not wanted

    """
    #==========================================================================
    Results
    #==========================================================================
    """
    pm = format_results(MDL, data["Z_max"]) # Format output
    zmodstats = MDL.stats(chain=-1)["zmod"] # Take last chain
    zn_avg = zmodstats["mean"]
    zn_l95 = zmodstats["95% HPD interval"][0]
    zn_u95 = zmodstats["95% HPD interval"][1]
    avg = data["Z_max"]*(zn_avg[0] + 1j*zn_avg[1]) # (In complex notation, de-normalized)
    l95 = data["Z_max"]*(zn_l95[0] + 1j*zn_l95[1]) # (In complex notation, de-normalized)
    u95 = data["Z_max"]*(zn_u95[0] + 1j*zn_u95[1]) # (In complex notation, de-normalized)
    fit = {"best": avg, "lo95": l95, "up95": u95} # Best fit dict with 95% HDP

    # Output
    return {"pymc_model": MDL, "params": pm, "data": data, "fit": fit, "SIP_model": model, "path": filename, "mcmc": mcmc, "model_type": {"log_min_tau":log_min_tau, "c_exp":c_exp, "decomp_polyn":decomp_poly, "cc_modes":cc_modes}}
    # End of inversion


#==============================================================================
"""
References:

Chen, Jinsong, Andreas Kemna, and Susan S. Hubbard. 2008. “A Comparison between
    Gauss-Newton and Markov-Chain Monte Carlo–based Methods for Inverting
    Spectral Induced-Polarization Data for Cole-Cole Parameters.” Geophysics
    73 (6): F247–59. doi:10.1190/1.2976115.
Dias, Carlos A. 2000. “Developments in a Model to Describe Low-Frequency
    Electrical Polarization of Rocks.” Geophysics 65 (2): 437–51.
    doi:10.1190/1.1444738.
Gamerman, Dani, and Hedibert F. Lopes. 2006. Markov Chain Monte Carlo:
    Stochastic Simulation for Bayesian Inference, Second Edition. CRC Press.
Ghorbani, A., C. Camerlynck, N. Florsch, P. Cosenza, and A. Revil. 2007.
    “Bayesian Inference of the Cole–Cole Parameters from Time- and Frequency-
    Domain Induced Polarization.” Geophysical Prospecting 55 (4): 589–605.
    doi:10.1111/j.1365-2478.2007.00627.x.
Hoff, Peter D. 2009. A First Course in Bayesian Statistical Methods. Springer
    Science & Business Media.
Keery, John, Andrew Binley, Ahmed Elshenawy, and Jeremy Clifford. 2012.
    “Markov-Chain Monte Carlo Estimation of Distributed Debye Relaxations in
    Spectral Induced Polarization.” Geophysics 77 (2): E159–70.
    doi:10.1190/geo2011-0244.1.
Nordsiek, Sven, and Andreas Weller. 2008. “A New Approach to Fitting Induced-
    Polarization Spectra.” Geophysics 73 (6): F235–45. doi:10.1190/1.2987412.
Pelton, W. H., W. R. Sill, and B. D. Smith. 1983. Interpretation of Complex
    Resistivity and Dielectric Data — Part 1. Vol 29. Geophysical Transactions.
Pelton, W. H., S. H. Ward, P. G. Hallof, W. R. Sill, and P. H. Nelson. 1978.
    “Mineral Discrimination and Removal of Inductive Coupling with
    Multifrequency IP.” Geophysics 43 (3): 588–609. doi:10.1190/1.1440839.
"""
#==============================================================================



# -*- coding: utf-8 -*-
"""
Created on Tue Apr 21 12:05:22 2015
@author:    charleslberube@gmail.com
            École Polytechnique Montréal
Copyright (c) 2015-2016 Charles L. Bérubé
"""
from __future__ import print_function

from future import standard_library
standard_library.install_aliases()
from builtins import str
from builtins import range

from bisip.models import mcmcinv
import bisip.invResults as iR
import pickle as pickle
import os

def save_object(obj, filename):
    with open(filename, 'wb') as output:
        pickle.dump(obj, output, -1)

#==============================================================================
""" 1.
    Model to use ?"""
# ex: model = "ColeCole", "Dias", "Debye", "Shin"
#model = "ColeCole"
#model = "Dias"
#model = "PDebye"
model = "PDecomp"
#model = "DDebye"
#model = "Shin"

#==============================================================================
""" 2.
    Markov-chain Monte-Carlo parameters ?"""
mcmc_p = {"adaptive"   : True,
          "nb_chain"   : 1,
          "nb_iter"    : 150000,
          "nb_burn"    : 130000,
          "thin"       : 1,
          "tune_inter" : 10000,
          "prop_scale" : 1.0,
          "verbose"    : False,
          "cov_inter"  : 5000,
          "cov_delay"  : 10000,
          }
sol = []

#for noise in [10, 5, 1]:
#for noise in range(1,11):
for noise in [1]:
#noise = 5
    adapt = True
#    adapt = False
    repeat = 1
    save_as = "%dmrad_%s_%d_MCMC_Solutions_Adaptive_%s.pkl" %(noise,model,repeat,str(adapt))
    #save_as = "%dmrad_%d_MCMC_Traces_Adaptive_%s.pkl" %(noise,repeat,str(adapt))
    #save_as = "tests.pkl"
    
    #==============================================================================
    """ 3.
        Paths to files ?"""
        
    reflist = os.listdir("/Users/Charles/Documents/SIP dat files")
    reflist = [x for x in reflist if not x.startswith('.')]
    reflist = [x for x in reflist if "AVG" in x]
    reflist = [x for x in reflist if "Reciprocals" in x]
    
    reflist = [reflist[0]]
    
    filename = ["/Users/Charles/Documents/SIP dat files/"+x for x in reflist]
    
    #==============================================================================
    """ 4.
        Number of headers to skip ?"""
    skip_header = 3
    
    #==============================================================================
    """ 5.
        Phase units in raw data file ?"""
    # {"rad" = radians}  {"mrad" = milliradians}  {"deg" = degrés}
    ph_units = "mrad"
    
    #==============================================================================
    # Call to the inversion function for every file
    for i, fn in enumerate(filename):
#    for i in range(repeat):
#        fn = filename[0]
        print('\nReading file:', fn, '(#%d/%d)' %(i+1,len(filename)))
        sol.append(mcmcinv(model, fn, mcmc=mcmc_p, headers=skip_header, ph_units=ph_units, decomp_poly=4, cc_modes=2, c_exp=1.0, log_min_tau=-3, keep_traces=False))
    
        """Plot fit and data ?"""
        if True:
            fig_fit = iR.plot_fit(sol[i], save=True)
    
        """Save results ?"""
        if True:
            iR.save_resul(sol[i])
    
        """Plot Debye relaxation time distribution ?"""
        if True:
            fig_RTD = iR.plot_rtd(sol[i], save=True, draw=False)
    
        """Print numerical results ?"""
        if False:
            iR.print_resul(sol[i])
    
        """Plot parameter histograms ?"""
        if True:
            fig_histo = iR.plot_histo(sol[i], save=True)
    
        if True:
            fig_trace = iR.plot_traces(sol[i], save=True)
    
        """Plot parameter summary and Gelman-Rubin convergence test ?"""
        if False:
            fig_kde = iR.plot_KDE(sol, "a0", "a1", save=False)


iR.merge_results(sol[0], [x.split(".")[0] for x in reflist])

# For further use in Python
saved_sol = [{key: value for key, value in list(s.items()) if key not in ["pymc_model"]} for s in sol]
#save_object(saved_sol, save_as)
#print "Solutions saved in list under", save_as

#==============================================================================
#===========================DATA FILE TEMPLATE=================================
#==============================================================================

"""
FORMAT DES DONNÉES
Mettre les données dans un fichier .csv, .txt, ou .dat
La séparation des colonnes par une virgule est importante
Les unités de la phase peuvent être des miliradians, radians ou degrés
Les unités de l'amplitude peuvent être des Ohm ou Ohm-m
(À spécifier dans la section plus haut, ici dans l'exemple skip_header = 1)
La colonne 'Courant (A)' n'est pas obligatoire
Les données doivent être sous le format suivant :
=============================================================================
Freq (Hz), Res (Ohm-m),  Phase (deg), dRes (Ohm-m), dPhase (deg), Current (A)
9.600e+03, 1.02676e+05, -1.29148e+02, 1.398449e+02, 1.361356e+00, 1.57806e-05
6.000e+03, 1.05713e+05, -9.52791e+01, 2.790833e+01, 2.635447e-01, 1.54739e-05
3.752e+03, 1.07753e+05, -7.10489e+01, 3.857578e+01, 3.577924e-01, 8.15439e-06
2.345e+03, 1.09314e+05, -5.39818e+01, 2.168789e+02, 1.984439e+00, 2.66668e-06
........., ..........., ............, ............, ............, ...........
........., ..........., ............, ............, ............, ...........
........., ..........., ............, ............, ............, ...........
7.584e-02, 1.23939e+05, -1.37357e+01, 3.042719e+02, 2.455678e+00, 8.20719e-06
4.740e-02, 1.24433e+05, -1.35053e+01, 4.442275e+01, 3.577924e-01, 7.94366e-06
2.962e-02, 1.24935e+05, -1.30672e+01, 4.100396e+02, 3.281218e+00, 6.95748e-06
=============================================================================
"""

#==============================================================================
#================================REFERENCES====================================
#==============================================================================
"""
Chen, Jinsong, Andreas Kemna, and Susan S. Hubbard. 2008. “A Comparison between
    Gauss-Newton and Markov-Chain Monte Carlo–based Methods for Inverting
    Spectral Induced-Polarization Data for Cole-Cole Parameters.” Geophysics
    73 (6): F247–59. doi:10.1190/1.2976115.
Dias, Carlos A. 2000. “Developments in a Model to Describe Low-Frequency
    Electrical Polarization of Rocks.” Geophysics 65 (2): 437–51.
    doi:10.1190/1.1444738.
Gamerman, Dani, and Hedibert F. Lopes. 2006. Markov Chain Monte Carlo:
    Stochastic Simulation for Bayesian Inference, Second Edition. CRC Press.
Ghorbani, A., C. Camerlynck, N. Florsch, P. Cosenza, and A. Revil. 2007.
    “Bayesian Inference of the Cole–Cole Parameters from Time- and Frequency-
    Domain Induced Polarization.” Geophysical Prospecting 55 (4): 589–605.
    doi:10.1111/j.1365-2478.2007.00627.x.
Hoff, Peter D. 2009. A First Course in Bayesian Statistical Methods. Springer
    Science & Business Media.
Keery, John, Andrew Binley, Ahmed Elshenawy, and Jeremy Clifford. 2012.
    “Markov-Chain Monte Carlo Estimation of Distributed Debye Relaxations in
    Spectral Induced Polarization.” Geophysics 77 (2): E159–70.
    doi:10.1190/geo2011-0244.1.
Nordsiek, Sven, and Andreas Weller. 2008. “A New Approach to Fitting Induced-
    Polarization Spectra.” Geophysics 73 (6): F235–45. doi:10.1190/1.2987412.
Patil, A., D. Huard and C.J. Fonnesbeck. 2010. PyMC: Bayesian Stochastic
    Modelling in Python. Journal of Statistical Software, 35(4), pp. 1-81
Pelton, W. H., W. R. Sill, and B. D. Smith. 1983. Interpretation of Complex
    Resistivity and Dielectric Data — Part 1. Vol 29. Geophysical Transactions.
Pelton, W. H., S. H. Ward, P. G. Hallof, W. R. Sill, and P. H. Nelson. 1978.
    “Mineral Discrimination and Removal of Inductive Coupling with
    Multifrequency IP.” Geophysics 43 (3): 588–609. doi:10.1190/1.1440839.
Shin, S. W., S. Park, and D. B. Shin. 2015. “Development of a New Equivalent
    Circuit Model for Spectral Induced Polarization Data Analysis of Ore
    Samples.” Environmental Earth Sciences 74 (7): 5711–16.
    doi:10.1007/s12665-015-4588-z."""


# -*- coding: utf-8 -*-
"""
Created on Sat Apr 25 14:51:26 2015

@author:    charleslberube@gmail.com
            École Polytechnique de Montréal

The MIT License (MIT)

Copyright (c) 2016 Charles L. Bérubé

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.

https://opensource.org/licenses/MIT
https://github.com/clberube/bisip

This Python module contains functions to visualize the Bayesian inversion results
"""
from __future__ import division
from __future__ import print_function

#==============================================================================
from builtins import zip
from builtins import str
from builtins import range
from past.builtins import basestring
from past.utils import old_div
import matplotlib.pyplot as plt
import matplotlib.gridspec as gridspec
import matplotlib.ticker as mticker
import numpy as np
from os import path, makedirs
from os import getcwd
from sys import argv
from math import ceil
from pymc import raftery_lewis, gelman_rubin, geweke
from scipy.stats import norm, gaussian_kde
from bisip.models import get_data
#==============================================================================

sym_labels = dict([('resi', r"$\rho\/(\Omega\cdot m)$"),
                   ('freq', r"Frequency $(Hz)$"),
                   ('phas', r"-Phase (mrad)"),
                   ('ampl', r"$|\rho|$ (normalized)"),
                   ('real', r"$\rho$' (normalized)"),
                   ('imag', r"$-\rho$'' (normalized)")])

def flatten(x):
    result = []
    for el in x:
        if hasattr(el, "__iter__") and not isinstance(el, basestring):
            result.extend(flatten(el))
        else:
            result.append(el)
    return result

def get_model_type(sol):
    model = sol["SIP_model"]
    if model == "PDecomp":
        if sol["model_type"]["c_exp"] == 0.5:
            model = "WarburgDecomp"
        elif sol["model_type"]["c_exp"] == 1.0:
            model = "DebyeDecomp"
        else:
            model = "ColeColeDecomp"
    return model

def print_resul(sol):
#==============================================================================
    # Impression des résultats
    pm, model, filename = sol["params"], sol["SIP_model"], sol["path"]
    print('\n\nInversion success!')
    print('Name of file:', filename)
    print('Model used:', model)
    e_keys = sorted([s for s in list(pm.keys()) if "_std" in s])
    v_keys = [e.replace("_std", "") for e in e_keys]
    labels = ["{:<8}".format(x+":") for x in v_keys]
    np.set_printoptions(formatter={'float': lambda x: format(x, '6.3E')})
    for l, v, e in zip(labels, v_keys, e_keys):
        print(l, pm[v], '+/-', pm[e], np.char.mod('(%.2f%%)',abs(100*pm[e]/pm[v])))

def plot_histo(sol, no_subplots=False, save=False, save_as_png=True):
    if save_as_png:
        save_as = 'png'
    else:
        save_as = 'pdf'
    MDL = sol["pymc_model"]
    model = get_model_type(sol)
    filename = sol["path"].replace("\\", "/").split("/")[-1].split(".")[0]
    keys = sorted([x.__name__ for x in MDL.deterministics]) + sorted([x.__name__ for x in MDL.stochastics])
    try:
        keys.remove("zmod")
        keys.remove("m_")

    except:
        pass
    for (i, k) in enumerate(keys):
        vect = old_div((MDL.trace(k)[:].size),(len(MDL.trace(k)[:])))
        if vect > 1:
            if "Decomp" in model:
                keys[i] = [k+"%d"%n for n in range(0,vect)]
            else:
                keys[i] = [k+"%d"%n for n in range(1,vect+1)]
    keys = list(flatten(keys))
    if no_subplots:
        figs = {}
        for c, k in enumerate(keys):
            fig, ax = plt.subplots(figsize=(6,4))
            if k == "R0":
                stoc = "R0"
            else:
                stoc =  ''.join([i for i in k if not i.isdigit()])
                stoc_num = [int(i) for i in k if i.isdigit()]
            try:
                data = sorted(MDL.trace(stoc)[:][:,stoc_num[0]-1])
            except:
                data = sorted(MDL.trace(stoc)[:])
            plt.xlabel("%s value"%k)
            plt.ylabel("Probability density")
            hist = plt.hist(data, bins=20, normed=True, linewidth=1.0, color="white")
            fit = norm.pdf(data, np.mean(data), np.std(data))
            plt.plot(data, fit, "-", label="Fitted PDF", linewidth=1.5)
            plt.legend(loc='best')
            plt.grid('off')
            if save:
                save_where = '/Figures/Histograms/%s/' %filename
                working_path = getcwd().replace("\\", "/")+"/"
                save_path = working_path+save_where
                if c == 0:
                    print("\nSaving histogram figures in:\n", save_path)
                if not path.exists(save_path):
                    makedirs(save_path)
                fig.savefig(save_path+'Histo-%s-%s-%s.%s'%(model,filename,k,save_as), bbox_inches='tight')
            figs[k] = fig
            plt.close(fig)
        return figs

    else:
        ncols = 2
        nrows = int(ceil(len(keys)*1.0 / ncols))
        fig, ax = plt.subplots(nrows, ncols, figsize=(7,nrows*1.8))
        for c, (a, k) in enumerate(zip(ax.flat, keys)):
            if k == "R0":
                stoc = "R0"
            else:
                stoc =  ''.join([i for i in k if not i.isdigit()])
                stoc_num = [int(i) for i in k if i.isdigit()]
            try:
                data = sorted(MDL.trace(stoc)[:][:,stoc_num[0]-1])
            except:
                data = sorted(MDL.trace(stoc)[:])
            plt.axes(a)
            plt.locator_params(axis = 'y', nbins = 8)
            plt.locator_params(axis = 'x', nbins = 7)
            plt.xlabel(k)
            try:
                hist = plt.hist(data, bins=20, normed=False, label=filename, edgecolor='#1f77b4', linewidth=1.0, color='#1f77b4', alpha=0.3)
                fit = norm.pdf(data, np.mean(data), np.std(data))                
                xh = [0.5 * (hist[1][r] + hist[1][r+1]) for r in range(len(hist[1])-1)]
                binwidth = old_div((max(xh) - min(xh)), len(hist[1]))
                fit *= len(data) * binwidth
                plt.plot(data, fit, "-", color='#ff7f0e', linewidth=1.5)
            except:
                print("File %s: could not plot %s histogram. Parameter is unstable (see trace)." %(filename,k))
            plt.grid('off')
            plt.ticklabel_format(style='sci', axis='both', scilimits=(0,0))
        
        for c in range(nrows):
            ax[c][0].set_ylabel("Frequency")

        plt.tight_layout(pad=1, w_pad=1, h_pad=0)
        for a in ax.flat[ax.size - 1:len(keys) - 1:-1]:
            a.set_visible(False)
        if save:
            save_where = '/Figures/Histograms/'
            working_path = getcwd().replace("\\", "/")+"/"
            save_path = working_path+save_where
            print("\nSaving parameter histograms in:\n", save_path)
            if not path.exists(save_path):
                makedirs(save_path)
            fig.savefig(save_path+'Histo-%s-%s.%s'%(model,filename,save_as), bbox_inches='tight')
        try:    plt.close(fig)
        except: pass
        return fig

def plot_KDE(sol, var1, var2, fig=None, ax=None, save=False, save_as_png=True):
    if True:
        save_as = 'png'
    else:
        save_as = 'pdf'
    if var1 == var2:
        fig, ax = plt.subplots(figsize=(3,3))
        plt.close(fig)
        return fig
    else:
        if fig == None or ax == None:
            fig, ax = plt.subplots(figsize=(3,3))
        MDL = sol["pymc_model"]
        filename = sol["path"].replace("\\", "/").split("/")[-1].split(".")[0]
        model = get_model_type(sol)
        if var1 == "R0":
            stoc1 = "R0"
        else:
            stoc1 =  ''.join([i for i in var1 if not i.isdigit()])
            stoc_num1 = [int(i) for i in var1 if i.isdigit()]
        try:
            x = MDL.trace(stoc1)[:,stoc_num1[0]-1]
        except:
            x = MDL.trace(stoc1)[:]
        if var2 == "R0":
            stoc2 = "R0"
        else:
            stoc2 =  ''.join([i for i in var2 if not i.isdigit()])
            stoc_num2 = [int(i) for i in var2 if i.isdigit()]
        try:
            y = MDL.trace(stoc2)[:,stoc_num2[0]-1]
        except:
            y = MDL.trace(stoc2)[:]
        xmin, xmax = min(x), max(x)
        ymin, ymax = min(y), max(y) 
        # Peform the kernel density estimate
        xx, yy = np.mgrid[xmin:xmax:100j, ymin:ymax:100j]
        positions = np.vstack([xx.ravel(), yy.ravel()])
        values = np.vstack([x, y])
        kernel = gaussian_kde(values)
        kernel.set_bandwidth(bw_method='silverman')
        kernel.set_bandwidth(bw_method=kernel.factor * 2.)
        f = np.reshape(kernel(positions).T, xx.shape)
    
        ax.set_xlim(xmin, xmax)
        ax.set_ylim(ymin, ymax)
        plt.axes(ax)
        # Contourf plot
        plt.grid(None)
        plt.ticklabel_format(style='sci', axis='both', scilimits=(0,0))
        ax.scatter(x, y, color='k', s=1)
        plt.xticks(rotation=90)
        plt.locator_params(axis = 'y', nbins = 7)
        plt.locator_params(axis = 'x', nbins = 7)
        cfset = ax.contourf(xx, yy, f, cmap=plt.cm.Blues, alpha=0.7)
        ## Or kernel density estimate plot instead of the contourf plot
#        ax.imshow(np.rot90(f), cmap='Blues', extent=[xmin, xmax, ymin, ymax])
        # Contour plot
#        cset = ax.contour(xx, yy, f, levels=cfset.levels[2::2], colors='k', alpha=0.8)
        # Label plot
    #    ax.clabel(cset, cset.levels[::1], inline=1, fmt='%.1E', fontsize=10)
        plt.yticks(fontsize=14)
        plt.xticks(fontsize=14)
        plt.ylabel("%s" %var2, fontsize=14)
        plt.xlabel("%s" %var1, fontsize=14)
    #    plt.gca().get_xaxis().get_major_formatter().set_useOffset(False)
    #    plt.gca().get_yaxis().get_major_formatter().set_useOffset(False)
        if save:
            save_where = '/Figures/Bivariate KDE/%s/' %filename
            working_path = getcwd().replace("\\", "/")+"/"
            save_path = working_path+save_where
            print("\nSaving KDE figure in:\n", save_path)
            if not path.exists(save_path):
                makedirs(save_path)
            fig.savefig(save_path+'KDE-%s-%s_%s_%s.%s'%(model,filename,var1,var2,save_as), dpi=200)
        plt.close(fig)
        return fig

def plot_all_KDE(sol):
    MDL = sol["pymc_model"]
    model = get_model_type(sol)
    filename = sol["path"].replace("\\", "/").split("/")[-1].split(".")[0]
    keys = sorted([x.__name__ for x in MDL.deterministics]) + sorted([x.__name__ for x in MDL.stochastics])
    sampler = MDL.get_state()["sampler"]
    try:
        keys.remove("zmod")
        keys.remove("m_")
        keys.remove("log_mean_tau")
        keys.remove("total_m")

    except:
        pass
    for (i, k) in enumerate(keys):
        vect = old_div((MDL.trace(k)[:].size),(len(MDL.trace(k)[:])))
        if vect > 1:
            keys[i] = [k+"%d"%n for n in range(0,vect)]
    keys = list(flatten(keys))
    ncols = len(keys)
    nrows = len(keys)
#    fig, ax = plt.subplots(nrows, ncols, figsize=(10,10))
    
    fig = plt.figure(figsize=(10,10))
#    plt.ticklabel_format(style='sci', axis='both', scilimits=(0,0))
    plotz = len(keys)    
    for i in range(plotz):
        for j in range(plotz):
            if j<i:
                var1 = keys[j]
                var2 = keys[i]
                print((var1, var2))
                ax = plt.subplot2grid((plotz-1, plotz-1), (i-1,j))
                ax.ticklabel_format(axis='y', style='sci', scilimits=(0,1))
                ax.ticklabel_format(axis='x', style='sci', scilimits=(0,1))
                
                if var1 == "R0":
                    stoc1 = "R0"
                else:
                    stoc1 =  ''.join([k for k in var1 if not k.isdigit()])
                    stoc_num1 = [int(k) for k in var1 if k.isdigit()]
                try:
                    x = MDL.trace(stoc1)[:,stoc_num1[0]-1]
                except:
                    x = MDL.trace(stoc1)[:]
                if var2 == "R0":
                    stoc2 = "R0"
                else:
                    stoc2 =  ''.join([k for k in var2 if not k.isdigit()])
                    stoc_num2 = [int(k) for k in var2 if k.isdigit()]
                try:
                    y = MDL.trace(stoc2)[:,stoc_num2[0]-1]
                except:
                    y = MDL.trace(stoc2)[:]
                xmin, xmax = min(x), max(x)
                ymin, ymax = min(y), max(y) 
                # Peform the kernel density estimate
                xx, yy = np.mgrid[xmin:xmax:100j, ymin:ymax:100j]
                positions = np.vstack([xx.ravel(), yy.ravel()])
                values = np.vstack([x, y])
                kernel = gaussian_kde(values)
                kernel.set_bandwidth(bw_method='silverman')
                kernel.set_bandwidth(bw_method=kernel.factor * 2.)
                f = np.reshape(kernel(positions).T, xx.shape)
            
                ax.set_xlim(xmin, xmax)
                ax.set_ylim(ymin, ymax)
                plt.axes(ax)
                # Contourf plot
                plt.grid(None)
                ax.scatter(x, y, color='k', s=1)
                plt.xticks(rotation=90)
                plt.locator_params(axis = 'y', nbins = 7)
                plt.locator_params(axis = 'x', nbins = 7)
                cfset = ax.contourf(xx, yy, f, cmap=plt.cm.Blues, alpha=0.7)
                ## Or kernel density estimate plot instead of the contourf plot
        #        ax.imshow(np.rot90(f), cmap='Blues', extent=[xmin, xmax, ymin, ymax])
                # Contour plot
        #        cset = ax.contour(xx, yy, f, levels=cfset.levels[2::2], colors='k', alpha=0.8)
                # Label plot
            #    ax.clabel(cset, cset.levels[::1], inline=1, fmt='%.1E', fontsize=10)
                plt.yticks(fontsize=14)
                plt.xticks(fontsize=14)
                if j == 0:
                    plt.ylabel("%s" %var2, fontsize=14)
                if i == len(keys)-1:
                    plt.xlabel("%s" %var1, fontsize=14)
                if j != 0:
                    ax.yaxis.set_ticklabels([])
                if i != len(keys)-1:
                    ax.xaxis.set_ticklabels([])
                plt.suptitle("Adaptive Metropolis step method", fontsize=14)
    
    
    
    
#    for v1 in range(len(keys)):
#        for v2 in range(len(keys)):
#            if v1 < v2:
#            
##                if v1 == v2:
##                    if keys[v1] == "R0":
##                        stoc1 = "R0"
##                    else:
##                        stoc1 =  ''.join([i for i in keys[v1] if not i.isdigit()])
##                        stoc_num1 = [int(i) for i in keys[v1] if i.isdigit()]
##                    try:
##                        x = MDL.trace(stoc1)[:,stoc_num1[0]-1]
##                    except:
##                        x = MDL.trace(stoc1)[:]
##                        ax[v1,v2].hist(x)
#                
#                fig = plot_KDE(sol, keys[v1], keys[v2], fig, ax[v2, v1])
#                print v1, v2
##            ax[i,j] = axs
#            fig = plot_axes(axs, fig)
#            fig.axes.append(axs)

    fig.tight_layout(pad=0, w_pad=1.0, h_pad=0.1)         
    return fig

def plot_hexbin(sol, var1, var2, save=False, save_as_png=True):
    if save_as_png:
        save_as = 'png'
    else:
        save_as = 'pdf'
    MDL = sol["pymc_model"]
    filename = sol["path"].replace("\\", "/").split("/")[-1].split(".")[0]
    model = get_model_type(sol)
    if var1 == "R0":
        stoc1 = "R0"
    else:
        stoc1 =  ''.join([i for i in var1 if not i.isdigit()])
        stoc_num1 = [int(i) for i in var1 if i.isdigit()]
    try:
        x = MDL.trace(stoc1)[:,stoc_num1[0]-1]
    except:
        x = MDL.trace(stoc1)[:]
    if var2 == "R0":
        stoc2 = "R0"
    else:
        stoc2 =  ''.join([i for i in var2 if not i.isdigit()])
        stoc_num2 = [int(i) for i in var2 if i.isdigit()]
    try:
        y = MDL.trace(stoc2)[:,stoc_num2[0]-1]
    except:
        y = MDL.trace(stoc2)[:]
    xmin, xmax = min(x), max(x)
    ymin, ymax = min(y), max(y)
    fig, ax = plt.subplots(figsize=(4,4))
    plt.grid(None)
    ax.set_xlim(xmin, xmax)
    ax.set_ylim(ymin, ymax)
#    plt.scatter(x, y)
    plt.hexbin(x, y, gridsize=20, cmap=plt.cm.Blues)
    plt.ticklabel_format(style='sci', axis='both', scilimits=(0,0))
    plt.xticks(rotation=90)
    plt.locator_params(axis = 'y', nbins = 5)
    plt.locator_params(axis = 'x', nbins = 5)    
    cb = plt.colorbar()
    cb.set_label('Number of observations')
    plt.yticks(fontsize=14)
    plt.xticks(fontsize=14)
    plt.ylabel("%s" %var2, fontsize=14)
    plt.xlabel("%s" %var1, fontsize=14)
    if save:
        save_where = '/Figures/Hexbins/%s/' %filename
        working_path = getcwd().replace("\\", "/")+"/"
        save_path = working_path+save_where
        print("\nSaving hexbin figure in:\n", save_path)
        if not path.exists(save_path):
            makedirs(save_path)
        fig.savefig(save_path+'Bivar-%s-%s_%s_%s.%s'%(model,filename,var1,var2,save_as))
    plt.close(fig)
    return fig

def plot_traces(sol, no_subplots=False, save=False, save_as_png=True):
    if save_as_png:
        save_as = 'png'
    else:
        save_as = 'pdf'
    MDL = sol["pymc_model"]
    model = get_model_type(sol)
    filename = sol["path"].replace("\\", "/").split("/")[-1].split(".")[0]
    keys = sorted([x.__name__ for x in MDL.deterministics]) + sorted([x.__name__ for x in MDL.stochastics])
    sampler = MDL.get_state()["sampler"]
    try:
        keys.remove("zmod")
        keys.remove("m_")
    except:
        pass
    for (i, k) in enumerate(keys):
        vect = old_div((MDL.trace(k)[:].size),(len(MDL.trace(k)[:])))
        if vect > 1:
            if "Decomp" in model:
                keys[i] = [k+"%d"%n for n in range(0,vect)]
            else:
                keys[i] = [k+"%d"%n for n in range(1,vect+1)]
    keys = list(flatten(keys))
    ncols = 2
    nrows = int(ceil(len(keys)*1.0 / ncols))
    if no_subplots:
        figs = {}
        for c, k in enumerate(keys):
            fig, ax = plt.subplots(figsize=(8,4))
            plt.ticklabel_format(style='sci', axis='both', scilimits=(0,0))
            if k == "R0":
                stoc = "R0"
            else:
                stoc =  ''.join([i for i in k if not i.isdigit()])
                stoc_num = [int(i) for i in k if i.isdigit()]
            try:
                data = MDL.trace(stoc)[:][:,stoc_num[0]-1]
            except:
                data = MDL.trace(stoc)[:]
            x = np.arange(sampler["_burn"]+1, sampler["_iter"]+1, sampler["_thin"])
            plt.ylabel("%s value" %k)
            plt.xlabel("Iteration number")
            plt.plot(x, data, '-', label=k, linewidth=2.0)
            plt.grid('on')
            if save:
                save_where = '/Figures/Traces/%s/' %filename
                working_path = getcwd().replace("\\", "/")+"/"
                save_path = working_path+save_where
                if c == 0:
                    print("\nSaving traces figure in:\n", save_path)
                if not path.exists(save_path):
                    makedirs(save_path)
                fig.savefig(save_path+'Trace-%s-%s-%s.%s'%(model,filename,k,save_as))
            figs[k] = fig
            plt.close(fig)
        return figs

    else:
        fig, ax = plt.subplots(nrows, ncols, figsize=(8,nrows*1.4))
        for c, (a, k) in enumerate(zip(ax.flat, keys)):
            if k == "R0":
                stoc = "R0"
            else:
                stoc =  ''.join([i for i in k if not i.isdigit()])
                stoc_num = [int(i) for i in k if i.isdigit()]
            try:
                data = MDL.trace(stoc)[:][:,stoc_num[0]-1]
            except:
                data = MDL.trace(stoc)[:]
            x = np.arange(sampler["_burn"]+1, sampler["_iter"]+1, sampler["_thin"])
            plt.axes(a)
            plt.ticklabel_format(style='sci', axis='both', scilimits=(0,0))
            plt.locator_params(axis = 'y', nbins = 6)
            plt.ylabel(k)
            plt.plot(x, data, '-', label=filename, linewidth=1.0)
            plt.grid('on')
            
        plt.tight_layout(pad=0.1, w_pad=0., h_pad=-2)
        for a in ax.flat[ax.size - 1:len(keys) - 1:-1]:
            a.set_visible(False)
        
        ax[:,0][-1].set_xlabel("Iteration number")
        for a in ax[:-1]:
            a[0].axes.get_xaxis().set_ticklabels([])
        
        if len(keys) % 2 == 0:
            ax[:,1][-1].set_xlabel("Iteration number")
            for a in ax[:-1]:
                a[1].axes.get_xaxis().set_ticklabels([])
        else:
            ax[:,1][-2].set_xlabel("Iteration number")
            for a in ax[:-2]:    
                a[1].axes.get_xaxis().set_ticklabels([])
            
        if save:
            save_where = '/Figures/Traces/'
            working_path = getcwd().replace("\\", "/")+"/"
            save_path = working_path+save_where
            print("\nSaving trace figures in:\n", save_path)
            if not path.exists(save_path):
                makedirs(save_path)
            fig.savefig(save_path+'Trace-%s-%s.%s'%(model,filename,save_as))
        plt.close(fig)
        return fig

def plot_summary(sol, save=False, save_as_png=True):
    if save_as_png:
        save_as = 'png'
    else:
        save_as = 'pdf'
    MDL, ch_n = sol["pymc_model"], sol["mcmc"]["nb_chain"]
    model = get_model_type(sol)
    filename = sol["path"].replace("\\", "/").split("/")[-1].split(".")[0]
    keys = sorted([x.__name__ for x in MDL.deterministics]) + sorted([x.__name__ for x in MDL.stochastics])
    try:
        keys.remove("zmod")
        keys.remove("m_")
    except:
        pass
    for (i, k) in enumerate(keys):
        vect = old_div((MDL.trace(k)[:].size),(len(MDL.trace(k)[:])))
        if vect > 1:
         keys[i] = [k+"%d"%n for n in range(1,vect+1)]
    keys = list(reversed(sorted(flatten(keys))))
    try:    r_hat = gelman_rubin(MDL)
    except:
        print("\nTwo or more chains of equal length required for Gelman-Rubin convergence")
    fig, axes = plt.subplots(figsize=(8,5))
    gs2 = gridspec.GridSpec(3, 3)
    ax1 = plt.subplot(gs2[:, :-1])
    ax2 = plt.subplot(gs2[:, -1], sharey = ax1)
    ax2.set_xlabel("R-hat")
    ax2.plot([1,1], [-1,len(keys)], "--b")
    for (i, k) in enumerate(keys):
        test = k[-1] not in ["%d"%d for d in range(1,8)] or k == "R0"
        for c in range(ch_n):
            if test:
                imp = None
                val_m = MDL.stats(k[:imp], chain=c)[k[:imp]]['mean']
                hpd_h = MDL.stats(k[:imp], chain=c)[k[:imp]]['95% HPD interval'][0]
                hpd_l = MDL.stats(k[:imp], chain=c)[k[:imp]]['95% HPD interval'][1]
            else:
                imp = -1
                val_m = MDL.stats(k[:imp], chain=c)[k[:imp]]['mean'][int(k[-1])-1]
                hpd_h = MDL.stats(k[:imp], chain=c)[k[:imp]]['95% HPD interval'][0][int(k[-1])-1]
                hpd_l = MDL.stats(k[:imp], chain=c)[k[:imp]]['95% HPD interval'][1][int(k[-1])-1]
            val = val_m
            err = [[abs(hpd_h-val_m)],
                    [abs(hpd_l-val_m)]]
            if ch_n % 2 != 0:   o_s = 0
            else:               o_s = 0.5
            ax1.scatter(val, i - (old_div(ch_n,2))*(1./ch_n/1.4) + (1./ch_n/1.4)*(c+o_s), color="DeepSkyBlue", marker="o", s=50, edgecolors='k')
            ax1.errorbar(val, i - (old_div(ch_n,2))*(1./ch_n/1.4) + (1./ch_n/1.4)*(c+o_s), xerr=err, color="k", fmt=" ", zorder=0)
        if ch_n >= 2:
            R = np.array(r_hat[k[:imp]])
            R[R > 3] = 3
            if test:
                ax2.scatter(R, i, color="b", marker="s", s=50, edgecolors='k')
            else:
                ax2.scatter(R[int(k[-1])-1], i, color="b", marker="s", s=50, edgecolors='k')
    
    ax1.set_ylim([-1, len(keys)])
    ax1.set_yticks(list(range(0,len(keys))))
    ax1.set_yticklabels(keys)
    plt.setp(ax2.get_yticklabels(), visible=False)
    ax2.set_xlim([0.5, 3.5])
    ax2.set_xticklabels(["","1","2","3+"])
    ax2.set_xticks([0.5, 1, 2, 3])
    ax1.set_xlabel("Parameter values")

    if save:
        save_where = '/Figures/Summaries/'
        working_path = getcwd().replace("\\", "/")+"/"
        save_path = working_path+save_where
        print("\nSaving summary figure in:\n", save_path)
        if not path.exists(save_path):
            makedirs(save_path)
        fig.savefig(save_path+'Summary-%s-%s.%s'%(model,filename,save_as))
    try:    plt.close(fig)
    except: pass

    return fig

def plot_autocorr(sol, save=False, save_as_png=True):
    if save_as_png:
        save_as = 'png'
    else:
        save_as = 'pdf'
    MDL = sol["pymc_model"]
    model = get_model_type(sol)
    filename = sol["path"].replace("\\", "/").split("/")[-1].split(".")[0]
    keys = sorted([x.__name__ for x in MDL.deterministics]) + sorted([x.__name__ for x in MDL.stochastics])
    try:
        keys.remove("zmod")
        keys.remove("m_")
    except:
        pass
    for (i, k) in enumerate(keys):
        vect = old_div((MDL.trace(k)[:].size),(len(MDL.trace(k)[:])))
        if vect > 1:
         keys[i] = [k+"%d"%n for n in range(1,vect+1)]
    keys = list(flatten(keys))
    ncols = 2
    nrows = int(ceil(len(keys)*1.0 / ncols))
    fig, ax = plt.subplots(nrows, ncols, figsize=(10,nrows*2))
    plt.ticklabel_format(style='sci', axis='both', scilimits=(0,0))
    for (a, k) in zip(ax.flat, keys):
        if k[-1] not in ["%d"%d for d in range(1,8)] or k =="R0":
            data = sorted(MDL.trace(k)[:].ravel())
        else:
            data = sorted(MDL.trace(k[:-1])[:][:,int(k[-1])-1].ravel())
        plt.axes(a)
        plt.gca().get_yaxis().get_major_formatter().set_useOffset(False)
        plt.gca().get_xaxis().get_major_formatter().set_useOffset(False)
        plt.yticks(fontsize=12)
        plt.xticks(fontsize=12)
        plt.ylabel(k, fontsize=12)
        to_thin = old_div(len(data),50)
        if to_thin != 0: plt.xlabel("Lags / %d"%to_thin, fontsize=12)
        else: plt.xlabel("Lags", fontsize=12)
        max_lags = None
        if len(data) > 50: data= data[::to_thin]
        plt.acorr(data, usevlines=True, maxlags=max_lags, detrend=plt.mlab.detrend_mean)
        plt.grid(None)
    fig.tight_layout()
    for a in ax.flat[ax.size - 1:len(keys) - 1:-1]:
        a.set_visible(False)
    if save:
        save_where = '/Figures/Autocorrelations/'
        working_path = getcwd().replace("\\", "/")+"/"
        save_path = working_path+save_where
        print("\nSaving autocorrelation figure in:\n", save_path)
        if not path.exists(save_path):
            makedirs(save_path)
        fig.savefig(save_path+'Autocorr-%s-%s.%s'%(model,filename,save_as))
    try:    plt.close(fig)
    except: pass
    return fig

def plot_rtd(sol, save=False, draw=False, save_as_png=True):
    if save_as_png:
        save_as = 'png'
    else:
        save_as = 'pdf'
    filename = sol["path"].replace("\\", "/").split("/")[-1].split(".")[0]
    model = get_model_type(sol)
    if draw or save:
        fig, ax = plt.subplots(figsize=(5,3))
        x = np.log10(sol["data"]["tau"])
        xmax = max(x)-1
        xmin = min(x)+1
        x = np.linspace(xmin, xmax,100)
        y = 100*np.sum([a*(x**i) for (i, a) in enumerate(sol["params"]["a"])], axis=0)
        pymc = sol["pymc_model"]
        cond = np.r_[True, y[1:] > y[:-1]] & np.r_[y[:-1] > y[1:], True]
        cond[0] = False
        peak = pymc.stats()["log_peak_tau"]['mean']
        peak_unc = pymc.stats()["log_peak_tau"]['standard deviation']
        try: plt.errorbar(peak,y[cond][0]+0.2*y[cond][0],None,peak_unc,"v",c="#d62728",label=r"$\tau_{peak}$")
        except: print("No peak detected")
        plt.vlines(pymc.stats()["log_mean_tau"]['mean'],0,max(y),color="#2ca02c",label=r"$\bar{\tau}$")
        plt.vlines(pymc.stats()["log_half_tau"]['mean'],0,max(y),color='#1f77b4',label=r"$\tau_{50}$")
        plt.errorbar(x, y, None, None, "-", color="#7f7f7f", linewidth=2, label="RTD")  
        inter = pymc.stats()["log_mean_tau"]['95% HPD interval']
        ax.axvspan(inter[0], inter[1], alpha=0.2, color="#2ca02c")
        inter = pymc.stats()["log_half_tau"]['95% HPD interval']
        ax.axvspan(inter[0], inter[1], alpha=0.2, color='#1f77b4')
        ax.fill_between(x, 0, y, where=x >= sol["model_type"]["log_min_tau"], color="#7f7f7f", alpha=0.2,label=r"$\Sigma m$")
#        ax.yaxis.set_major_formatter(mticker.FormatStrFormatter('%.2f'))
        plt.xlabel(r"$log_{10}\tau$ ($\tau$ in s)", fontsize=12)
        plt.ylabel("Chargeability (%)", fontsize=12)
        plt.yticks(fontsize=12), plt.xticks(fontsize=12)
        plt.xlim([xmin, xmax])
        plt.grid('on')
        plt.ylim([0, max(y)])
        plt.legend(numpoints=1, fontsize=10, loc="best",labelspacing=0.1)
        fig.tight_layout()
    if save:
        save_where = '/Figures/Debye distributions/'
        working_path = getcwd().replace("\\", "/")+"/"

        save_path = working_path+save_where
        print("\nSaving relaxation time distribution figure in:\n", save_path)
        if not path.exists(save_path):
            makedirs(save_path)
        fig.savefig(save_path+'Polynomial-RTD-%s-%s.%s'%(model, filename,save_as))
    try:    plt.close(fig)
    except: pass
    if draw:    return fig
    else:       return None

def save_resul(sol):
    # Fonction pour enregistrer les résultats
    MDL, pm, filepath = sol["pymc_model"], sol["params"], sol["path"]
    model = get_model_type(sol)
    sample_name = filepath.replace("\\", "/").split("/")[-1].split(".")[0]
    save_where = '/Results/'
    working_path = getcwd().replace("\\", "/")+"/"

    save_path = working_path+save_where+"%s/"%sample_name
    print("\nSaving csv file in:\n", save_path)
    if not path.exists(save_path):
        makedirs(save_path)
    if model == 'Debye': tag = 0
    else: tag = 1
    A = []
    headers = []
    for c, key in enumerate(sorted(pm.keys())):
        A.append(list(np.array(pm[key]).ravel()))
        key = model[:2]+"_"+key
        if len(A[c]) == 1:
            headers.append(key)
        else:
            for i in range(len(A[c])):
                headers.append(key+"%d" %(i+tag))
    headers = ','.join(headers)
    results = np.array(flatten(A))
    if model in ["DDebye", "PDecomp"]:
        tau_ = sol["data"]["tau"]
        add = ["tau"+"%d"%(i+1) for i in range(len(tau_))]
        add = ',' + ','.join(add)
        headers += add
        results = np.concatenate((results,tau_))
    np.savetxt(save_path+'INV_%s_%s.csv' %(model,sample_name), results[None],
               header=headers, comments='', delimiter=',')
    vars_ = ["%s"%x for x in MDL.stochastics]+["%s"%x for x in MDL.deterministics]
    if "zmod" in vars_: vars_.remove("zmod")
    MDL.write_csv(save_path+'STATS_%s_%s.csv' %(model,sample_name), variables=(vars_))

def merge_results(sol,files):
    model = get_model_type(sol)
    save_where = '/Batch results/'
    working_path = getcwd().replace("\\", "/")+"/"
    save_path = working_path+save_where
    print("\nMerging csv files")
    if not path.exists(save_path):
        makedirs(save_path)
    to_merge = working_path+"/Results/%s/INV_%s_%s.csv" %(files[0],model,files[0])
    headers = np.genfromtxt(to_merge, delimiter=",", dtype=str, skip_footer=1)
    merged_inv_results = np.empty((len(files), len(headers)))
    for i, f in enumerate(files):
        merged_inv_results[i] = np.loadtxt(working_path+"/Results/%s/INV_%s_%s.csv" %(f,model,f), delimiter=",", skiprows=1)
    rows = np.array(files, dtype=str)[:, np.newaxis]
    hd = ",".join(["ID"] + list(headers))
    np.savetxt(save_path+"Merged_%s_%s_TO_%s.csv" %(model,files[0],files[-1]), np.hstack((rows, merged_inv_results)), delimiter=",", header=hd, fmt="%s")
    print("Batch file successfully saved in:\n", save_path)

def plot_data(filename, headers, ph_units):
    data = get_data(filename,headers,ph_units)
    # Graphiques du data
    Z = data["Z"]
    dZ = data["Z_err"]
    f = data["freq"]
    Zr0 = max(abs(Z))
    zn_dat = old_div(Z,Zr0)
    zn_err = old_div(dZ,Zr0)
    Pha_dat = 1000*data["pha"]
    Pha_err = 1000*data["pha_err"]
    Amp_dat = old_div(data["amp"],Zr0)
    Amp_err = old_div(data["amp_err"],Zr0)

    fig, ax = plt.subplots(3, 1, figsize=(6,8))
    for t in ax:
        t.tick_params(labelsize=12)
    # Real-Imag
    plt.axes(ax[0])
    plt.errorbar(zn_dat.real, -zn_dat.imag, zn_err.imag, zn_err.real, '.b', label=filename)
    plt.xlabel(sym_labels['real'], fontsize=12)
    plt.ylabel(sym_labels['imag'], fontsize=12)

    plt.xlim([None, 1])
    plt.ylim([0, None])
#    plt.legend(numpoints=1, fontsize=9)
#    plt.title(filename, fontsize=10)
    # Freq-Phas
    plt.axes(ax[1])
    plt.errorbar(f, -Pha_dat, Pha_err, None, '.b', label=filename)
    ax[1].set_yscale("log", nonposy='clip')
    ax[1].set_xscale("log")
    plt.xlabel(sym_labels['freq'], fontsize=12)
    plt.ylabel(sym_labels['phas'], fontsize=12)
#    plt.legend(loc=2, numpoints=1, fontsize=9)
    plt.ylim([1,1000])
    # Freq-Ampl
    plt.axes(ax[2])
    plt.errorbar(f, Amp_dat, Amp_err, None, '.b', label=filename)
    ax[2].set_xscale("log")
    plt.xlabel(sym_labels['freq'], fontsize=12)
    plt.ylabel(sym_labels['ampl'], fontsize=12)
    plt.ylim([None,1.0])
#    plt.legend(numpoints=1, fontsize=9)
    fig.tight_layout()

    plt.close(fig)
    return fig

def plot_deviance(sol, save=False, draw=True, save_as_png=True):
    if save_as_png:
        save_as = 'png'
    else:
        save_as = 'pdf'
    filename = sol["path"].replace("\\", "/").split("/")[-1].split(".")[0]
    model = get_model_type(sol)
    if draw or save:
        fig, ax = plt.subplots(figsize=(6,4))
        deviance = sol["pymc_model"].trace('deviance')[:]
        sampler_state = sol["pymc_model"].get_state()["sampler"]
        x = np.arange(sampler_state["_burn"]+1, sampler_state["_iter"]+1, sampler_state["_thin"])
        plt.plot(x, deviance, "-b", linewidth=2, label="Model deviance\nDIC = %.2f\nBPIC = %.2f" %(sol["pymc_model"].DIC,sol["pymc_model"].BPIC))
        plt.xlabel("Iteration", fontsize=14)
        plt.ylabel("Deviance", fontsize=14)
        plt.yticks(fontsize=14), plt.xticks(fontsize=14)
        plt.legend(numpoints=1, fontsize=14, loc="best")
        fig.tight_layout()
    if save:
        save_where = '/Figures/ModelDeviance/'
        working_path = getcwd().replace("\\", "/")+"/"
        save_path = working_path+save_where
        print("\nSaving model deviance figure in:\n", save_path)
        if not path.exists(save_path):
            makedirs(save_path)
        fig.savefig(save_path+'ModelDeviance-%s-%s.%s'%(model,filename,save_as))
    try:    plt.close(fig)
    except: pass
    if draw:    return fig
    else:       return None

def logp_trace(model):
    """
    return a trace of logp for model
    """
    #init
    db = model.db
    n_samples = db.trace('deviance').length()
    logp = np.empty(n_samples, np.double)
    #loop over all samples
    for i_sample in range(n_samples):
        #set the value of all stochastic to their 'i_sample' value
        for stochastic in model.stochastics:
            try:
                value = db.trace(stochastic.__name__)[i_sample]
                stochastic.value = value

            except KeyError:
                print("No trace available for %s. " % stochastic.__name__)

        #get logp
        logp[i_sample] = model.logp
    return logp

def plot_logp(sol, save=False, draw=True, save_as_png=True):
    if save_as_png:
        save_as = 'png'
    else:
        save_as = 'pdf'
    filename = sol["path"].replace("\\", "/").split("/")[-1].split(".")[0]
    model = get_model_type(sol)
    if draw or save:
        fig, ax = plt.subplots(figsize=(6,4))
        logp = logp_trace(sol["pymc_model"])
        sampler_state = sol["pymc_model"].get_state()["sampler"]
        x = np.arange(sampler_state["_burn"]+1, sampler_state["_iter"]+1, sampler_state["_thin"])
        plt.plot(x, logp, "-b", linewidth=2, label="logp")
        plt.xlabel("Iteration", fontsize=14)
        plt.ylabel("Log-likelihood", fontsize=14)
        plt.yticks(fontsize=14), plt.xticks(fontsize=14)
        plt.legend(numpoints=1, fontsize=14, loc="best")
        fig.tight_layout()
    if save:
        save_where = '/Figures/LogLikelihood/'
        working_path = getcwd().replace("\\", "/")+"/"
        save_path = working_path+save_where
        print("\nSaving logp trace figure in:\n", save_path)
        if not path.exists(save_path):
            makedirs(save_path)
        fig.savefig(save_path+'LogLikelihood-%s-%s.%s'%(model,filename,save_as))
    try:    plt.close(fig)
    except: pass
    if draw:    return fig
    else:       return None

def plot_fit(sol, save=False, draw=True, save_as_png=True):
    if save_as_png:
        save_as = 'png'
    else:
        save_as = 'pdf'
    filepath = sol["path"]
    sample_name = filepath.replace("\\", "/").split("/")[-1].split(".")[0]
    model = get_model_type(sol)
    data = sol["data"]
    fit = sol["fit"]
    # Graphiques du fit
    f = data["freq"]
    Zr0 = max(abs(data["Z"]))
    zn_dat = old_div(data["Z"],Zr0)
    zn_err = old_div(data["Z_err"],Zr0)
    zn_fit = old_div(fit["best"],Zr0)
    zn_min = old_div(fit["lo95"],Zr0)
    zn_max = old_div(fit["up95"],Zr0)
    Pha_dat = 1000*data["pha"]
    Pha_err = 1000*data["pha_err"]
    Pha_fit = 1000*np.angle(fit["best"])
    Pha_min = 1000*np.angle(fit["lo95"])
    Pha_max = 1000*np.angle(fit["up95"])
    Amp_dat = old_div(data["amp"],Zr0)
    Amp_err = old_div(data["amp_err"],Zr0)
    Amp_fit = old_div(abs(fit["best"]),Zr0)
    Amp_min = old_div(abs(fit["lo95"]),Zr0)
    Amp_max = old_div(abs(fit["up95"]),Zr0)
    if draw or save:
        fig, ax = plt.subplots(1, 3, figsize=(11,3))
#        for t in ax:
#            t.tick_params(labelsize=14)
        # Real-Imag
        plt.axes(ax[2])
        plt.errorbar(zn_fit.real, -zn_dat.imag, zn_err.imag, zn_err.real, '.', label='Data')
        plt.plot(zn_fit.real, -zn_fit.imag, '-', label='Fitted model')
        plt.fill_between(zn_fit.real, -zn_max.imag, -zn_min.imag, color='#7f7f7f', alpha=0.2)
        plt.xlabel(sym_labels['real'])
        plt.ylabel(sym_labels['imag'])
        plt.legend(loc='best', fontsize=9, numpoints=1)
        plt.xlim([None, 1])
        plt.ylim([0, max(-zn_dat.imag)])
        
        # Freq-Ampl
        plt.axes(ax[1])
        plt.errorbar(f, Amp_dat, Amp_err, None, '.', label='Data')
        plt.semilogx(f, Amp_fit, '-', label='Fitted model')
        plt.fill_between(f, Amp_max, Amp_min, color='#7f7f7f', alpha=0.2)
        plt.xlabel(sym_labels['freq'])
        plt.ylabel(sym_labels['ampl'])
        plt.legend(loc='best', fontsize=9, numpoints=1)
        plt.xlim([10**np.floor(min(np.log10(f))), 10**np.ceil(max(np.log10(f)))])
        plt.ylim([None,1.0])

        # Freq-Phas
        plt.axes(ax[0])
        plt.errorbar(f, -Pha_dat, Pha_err, None, '.', label='Data')
        plt.loglog(f, -Pha_fit, '-', label='Fitted model')
        ax[0].set_yscale("log", nonposy='clip')
        plt.fill_between(f, -Pha_max, -Pha_min, color='#7f7f7f', alpha=0.2)
        plt.xlabel(sym_labels['freq'])
        plt.ylabel(sym_labels['phas'])
        plt.legend(loc='best', fontsize=9, numpoints=1)
        plt.xlim([10**np.floor(min(np.log10(f))), 10**np.ceil(max(np.log10(f)))])
        plt.ylim([1,10**np.ceil(max(np.log10(-Pha_dat)))])

        for a in ax:
            a.grid('on')

        if  (-Pha_dat < 1).any() and (-Pha_dat >= 0.1).any():
            plt.ylim([0.1,10**np.ceil(max(np.log10(-Pha_dat)))])  
        if  (-Pha_dat < 0.1).any() and (-Pha_dat >= 0.01).any():
            plt.ylim([0.01,10**np.ceil(max(np.log10(-Pha_dat)))]) 
        
        plt.tight_layout(pad=1, h_pad=0, w_pad=0)
        
    if save:
        save_where = '/Figures/Fit figures/'
        working_path = getcwd().replace("\\", "/")+"/"
        save_path = working_path+save_where
        print("\nSaving fit figure in:\n", save_path)
        if not path.exists(save_path):
            makedirs(save_path)
        fig.savefig(save_path+'FIT-%s-%s.%s'%(model,sample_name,save_as), bbox_inches='tight')
    try:    plt.close(fig)
    except: pass
    if draw:    return fig
    else:       return None

def print_diagn(M, q, r, s):
    return raftery_lewis(M, q, r, s, verbose=0)

def plot_par():
    from matplotlib.pyplot import rcParams
#    rc = rcParams.items()    
    rc = {u'figure.dpi': 72.0,
          u'figure.edgecolor': 'white',
          u'figure.facecolor': 'white',
          u'savefig.bbox': u'tight',
          u'savefig.directory': u'~',
          u'savefig.dpi': 200.0,
          u'savefig.edgecolor': u'white',
          u'savefig.facecolor': u'white',
          }
    return rc



# -*- coding: utf-8 -*-
#!python
#cython: boundscheck=False
#cython: cdivision=True
#cython: nonecheck=False
#cython: wraparound=False
#cython: boundscheck=False

"""
Created on Wed Nov  4 14:05:35 2015

@author:    clafreniereberube@gmail.com
            École Polytechnique de Montréal

Copyright (c) 2015-2016 Charles L. Bérubé

"""

import numpy as np
cimport numpy as cnp

from libc.stdlib cimport malloc, free

DTYPE = np.float_
ctypedef cnp.float_t DTYPE_t
DTYPE2 = np.complex128
ctypedef cnp.complex128_t DTYPE2_t

cdef double complex C_ColeCole(double w_, double m_, double lt_, double c_):
    cdef double complex Z_i
    cdef double complex jay
    jay.real = 0.0
    jay.imag = 1.0
    Z_i = m_*(1.0 - 1.0/(1.0 + ((jay*w_*(10.0**lt_))**c_)))
    return Z_i

cdef double complex C_Dias(double w_, double R0_, double m_, double log_tau_, double eta_,double delta_):
    cdef double complex Z_i
    cdef double complex jay
    jay.real = 0.0
    jay.imag = 1.0
    Z_i = R0_*(1 - (m_*(1 - 1.0/(1 + (jay*w_*(((10**log_tau_)/delta_)*(1 - delta_)/(1 - m_)))*(1 + 1.0/(jay*w_*(10**log_tau_) + eta_*(10**log_tau_)*(jay*w_**0.5)))))))
    return Z_i

cdef double complex C_Shin(double w_, double R_, double log_Q_, double n_):
    cdef double complex Z_i
    cdef double complex jay
    jay.real = 0.0
    jay.imag = 1.0
    Z_i = R_ / ((jay*w_**n_)*((10**log_Q_))*R_ + 1)
    return Z_i

cdef double complex C_Debye(double w_, double m_, double tau_, double c_):
    cdef double complex Z_i
    cdef double complex jay
    jay.real = 0.0
    jay.imag = 1.0
    Z_i = m_*(1 - 1.0/(1 + ((jay*w_*(tau_))**c_)))
    return Z_i

def ColeCole_cyth1(cnp.ndarray[DTYPE_t, ndim=1] w, DTYPE_t R0, cnp.ndarray[DTYPE_t, ndim=1] m, cnp.ndarray[DTYPE_t, ndim=1] lt, cnp.ndarray[DTYPE_t, ndim=1] c):
    cdef int N = w.shape[0]
    cdef int D = m.shape[0]
    cdef int i, j
    cdef double complex z_
#    cdef double zr = z_.real
#    cdef double zi = z_.imag
    cdef cnp.ndarray[DTYPE_t, ndim=2] Z = np.empty((2,N), dtype=DTYPE)
    for j in range(N):
        z_ = 0
        for i in range(D):
            z_ += C_ColeCole(w[j], m[i], lt[i], c[i])
        z_ = R0*(1 - z_)
        Z[0,j] = z_.real
        Z[1,j] = z_.imag
    return Z

def ColeCole_cyth2(double[:] w, double R0, double[:] m, double[:] lt, double[:] c):
    cdef int N = w.shape[0]
    cdef int D = m.shape[0]
    cdef int i, j
    cdef int rows = 2
    cdef double complex z_
    cdef double[:,:] Z = np.empty((rows,N), dtype=DTYPE)
    for j in range(N):
        z_ = 0
        for i in range(D):
            z_ += C_ColeCole(w[j], m[i], lt[i], c[i])
        z_ = R0*(1.0 - z_)
        Z[0,j] = z_.real
        Z[1,j] = z_.imag
    return Z

def Dias_cyth(cnp.ndarray[DTYPE_t, ndim=1] w, DTYPE_t R0, DTYPE_t m, DTYPE_t log_tau, DTYPE_t eta, DTYPE_t delta):
    cdef int N = w.shape[0]
    cdef int j
    cdef cnp.ndarray[DTYPE_t, ndim=2] Z = np.empty((2,N), dtype=DTYPE)
    cdef cnp.ndarray[DTYPE2_t, ndim=1] z_ = np.zeros(N, dtype=DTYPE2)
    for j in range(N):
        z_[j] = C_Dias(w[j], R0, m, log_tau, eta, delta)
        Z[0,j] = z_[j].real
        Z[1,j] = z_[j].imag
    return Z

def Decomp_cyth(cnp.ndarray[DTYPE_t, ndim=1] w, cnp.ndarray[DTYPE_t, ndim=1] tau_10, cnp.ndarray[DTYPE_t, ndim=2] log_taus, DTYPE_t c_exp, DTYPE_t R0, cnp.ndarray[DTYPE_t, ndim=1] a):
    cdef int D = a.shape[0]
    cdef int N = w.shape[0]
    cdef int S = tau_10.shape[0]
    cdef int i, j, k
    cdef cnp.ndarray[DTYPE_t, ndim=1] M = np.zeros(S, dtype=DTYPE)
    cdef double complex z_
    cdef double complex z_hi
    cdef cnp.ndarray[DTYPE_t, ndim=2] Z = np.empty((2,N), dtype=DTYPE)
    for i in range(D):
        for k in range(S):
            M[k] = M[k] + a[i]*(log_taus[i,k])    
    for j in range(N):
        z_ = 0 
        for k in range(S):
            z_ +=  C_Debye(w[j], M[k], tau_10[k], c_exp)
#        z_hi = C_Debye(w[j], m_hi, 10**log_tau_hi, 1.0)
#        z_ += z_hi
        z_ = R0*(1 - z_)
        Z[0,j] = z_.real
        Z[1,j] = z_.imag
    return Z

def Shin_cyth(cnp.ndarray[DTYPE_t, ndim=1] w, cnp.ndarray[DTYPE_t, ndim=1] R, cnp.ndarray[DTYPE_t, ndim=1] log_Q, cnp.ndarray[DTYPE_t, ndim=1] n):
    cdef int D = R.shape[0]
    cdef int N = w.shape[0]
    cdef int i, j
    cdef cnp.ndarray[DTYPE2_t, ndim=1] z_ = np.zeros(N, dtype=DTYPE2)
    cdef cnp.ndarray[DTYPE_t, ndim=2] Z = np.empty((2,N), dtype=DTYPE)
    for j in range(N):
        for i in range(D):
            z_[j] = z_[j] + C_Shin(w[j], R[i], log_Q[i], n[i])
        Z[0,j] = z_[j].real
        Z[1,j] = z_[j].imag
    return Z


# -*- coding: utf-8 -*-

"""
Created on Tue Apr 21 12:05:22 2015

@author:    charleslberube@gmail.com
            École Polytechnique de Montréal

The MIT License (MIT)

Copyright (c) 2015-2016 Charles L. Bérubé

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.

https://opensource.org/licenses/MIT
https://github.com/clberube/bisip

This python script builds the graphical user interface may be used to call the
Bayesian inversion module of all SIP models (BISIP_models.py)
"""

from __future__ import division
from __future__ import print_function
#from __future__ import unicode_literals

#==============================================================================
# System imports

from sys import version_info
print("Running Python %d.%d.%d"%version_info[0:3])

print("\nFuture imports")
from future import standard_library
standard_library.install_aliases()
from builtins import zip
from builtins import str
from builtins import map
from builtins import range
from past.builtins import basestring
from builtins import object
from past.utils import old_div

print("System imports")
from sys import argv, stdout
stdout.flush()
from platform import system
from os.path import realpath as osp_realpath
from os import getcwd
from json import load as jload, dump as jdump
from warnings import filterwarnings
filterwarnings('ignore') # Ignore some tkinter warnings

print("GUI imports")
if version_info[0] < 3:
    import Tkinter as tk
    import FixTk # To avoid pyinstaller error
else:
    import tkinter as tk
import tkinter.filedialog, tkinter.messagebox, tkinter.font

print("BISIP imports")
from bisip.models import mcmcinv
import bisip.invResults as iR

print("Other imports")
from matplotlib.backends.backend_tkagg import FigureCanvasTkAgg, NavigationToolbar2TkAgg
from matplotlib.pyplot import rcParams
from itertools import combinations
from collections import OrderedDict

print("All modules successfully loaded")
stdout.flush()

#==============================================================================
# Fonts
if "Darwin" in system():
    print("\nOS X detected")
    fontsize = 12
    pad_radio = 3
    but_size = -2
    res_size = -1
else:
    print("\nWindows detected")
    fontsize = 10
    pad_radio = 0
    but_size = -2
    res_size = -2

window_font = "TkDefaultFont %s"%fontsize
fontz = {"bold": ("TkDefaultFont", fontsize, "bold"),
         "normal_small": ("TkDefaultFont", fontsize+but_size, "normal"),
         "italic_small": ("TkDefaultFont", fontsize+but_size, "italic")}

# To flatten ND lists
def flatten(x):
    result = []
    for el in x:
        if hasattr(el, "__iter__") and not isinstance(el, basestring):
            result.extend(flatten(el))
        else:
            result.append(el)
    return result

#==============================================================================
# Import last used window parameters
# All GUI options and choices saved when closing the main window
# Parameters are saved in root_ini file in the executable's directory
# To reset and use default parameters, delete root_ini in local directory

class MainApplication(object):
#    working_path = str(osp_realpath(argv[0])).replace("\\", "/")+"/"
    working_path = getcwd().replace("\\", "/")+"/"
    def __init__(self, master, fontz):
        self.save_options = {"Save all hexbins (will make error)":            tk.BooleanVar(),
                             "Save all bivariate KDE (will make error)":      tk.BooleanVar(),
                             "Save fit figures":            tk.BooleanVar(),
                             "Save traces":                 tk.BooleanVar(),
                             "Save histograms":             tk.BooleanVar(),
                             "Save autocorrelations":       tk.BooleanVar(),
                             "Save Debye RTD":              tk.BooleanVar(),
                             "Save summaries":              tk.BooleanVar(),
                             "Save deviance":               tk.BooleanVar(),
                             "Save loglikelihood":          tk.BooleanVar(),
                             "Save traces as txt":          tk.BooleanVar(),
                             "Tuning verbose":              tk.BooleanVar(),
                             "No subplots":                 tk.BooleanVar(),
                             "PNG figures":                 tk.BooleanVar(),
                            }
        self.run_options = OrderedDict((
                            ("Auto draw fit",                tk.BooleanVar()),
                            ("Print results in console",     tk.BooleanVar()),
                            ("Save CSV results",             tk.BooleanVar()),
                            ))

        self.mcmc_vars = OrderedDict((
                                ("Number of chains"     , (tk.IntVar(), 1)),
                                ("Total iterations"     , (tk.IntVar(), 100000)),
                                ("Burn-in period"       , (tk.IntVar(), 80000)),
                                ("Thinning factor"      , (tk.IntVar(), 1)),
                                ("Tuning interval"      , (tk.IntVar(), 10000)),
                                ("Proposal scale"       , (tk.DoubleVar(), 1)),
                                ("Covariance delay"     , (tk.IntVar(), 10000)),
                                ("Covariance interval"  , (tk.IntVar(), 10000)),
                                ))

        self.master = master
        self.master.resizable(width=tk.FALSE, height=tk.FALSE)
        self.load()
        self.set_plot_par()
        self.build_helpmenu()
        self.make_main_frames()
        self.make_browse_button()
        self.headers_input()
        self.open_files = self.root_ini["Imported files"]
        self.draw_file_list()
        self.phase_units()
        self.mcmc_parameters()
        self.run_exit()
        self.make_options(from_root=True)
        self.model_choice()
        self.activity(idle=True)

    def use_default_root_ini(self):
        self.default_root = {
                            "Spectral IP model"     : "ColeCole",
                            "Adaptive Metropolis"   : True,
                            "Nb header lines"       : 1,
                            "Phase units"           : "mrad",
                            "Imported files"        : [],
                            "Polyn order"           : 4,
                            "Freq dep"              : 1.0,
                            "Nb modes"              : 2,
                            }
        for k, v in list(self.save_options.items()):
            self.default_root[k] = v.get()
        for k, v in list(self.mcmc_vars.items()):
            self.default_root[k] = v[1]
        for k, v in list(self.run_options.items()):
            self.default_root[k] = v.get()
        return self.default_root

#==============================================================================
# Main frames
#==============================================================================
    def make_main_frames(self):
        # Frame for importing files
        self.frame_import = tk.LabelFrame(self.master, text="1. Import data", width=200, height=5, font=fontz["bold"])
        self.frame_import.grid(row = 0, column=1, columnspan=1, rowspan=5, sticky=tk.W+tk.E+tk.N, padx=10, pady=(5,15))
        self.frame_import.grid_columnconfigure(0, weight=1), self.frame_import.grid_rowconfigure(0, weight=1)
        # Frame for choosing the model
        self.frame_model = tk.LabelFrame(self.master, text="2. SIP model", width=200, height=4, font=fontz["bold"])
        self.frame_model.grid(row = 6, column=1, columnspan=1, sticky=tk.W+tk.E+tk.N, padx=10, pady=(5,15))
        self.frame_model.grid_columnconfigure(0, weight=1), self.frame_model.grid_rowconfigure(0, weight=1)
        # Frame to enter mcmc parameters
        self.frame_mcmc = tk.LabelFrame(self.master, text="3. MCMC settings", width=200, height=4, font=fontz["bold"])
        self.frame_mcmc.grid(row = 7, column=1, columnspan=1, sticky=tk.W+tk.E+tk.N, padx=10, pady=(5,15), ipady=3)
        self.frame_mcmc.grid_columnconfigure(0, weight=1)
        # Frame to run and exit
        self.frame_ruex = tk.LabelFrame(self.master, text="4. Run", width=200, height=4, font=fontz["bold"])
        self.frame_ruex.grid(row = 8, column=1, columnspan=1, sticky=tk.S+tk.W+tk.E+tk.N, padx=10, pady=(5,10))
        self.frame_ruex.columnconfigure(0, weight=1)
        # Frame to list the imported files and preview
        self.frame_list = tk.LabelFrame(self.master, text="List of imported files", font=fontz["bold"])
        self.frame_list.grid(row = 0, column=0, columnspan=1, rowspan=15, sticky=tk.W+tk.E+tk.N+tk.S, padx=10, pady=(5,10))
        self.frame_list.grid_rowconfigure(1, weight=1), self.frame_list.columnconfigure(0, weight=1)
        # Frame to visualize results
        self.frame_results = tk.LabelFrame(self.master, text="Results", font=fontz["bold"])
        self.frame_results.grid(row = 0, column=2, columnspan=1, rowspan=15, sticky=tk.W+tk.E+tk.N+tk.S, padx=10, pady=(5,10))
        self.frame_results.grid_rowconfigure(14, weight=1)

#==============================================================================
# Running inversion
#==============================================================================
    def run_inversion(self):
        try:    self.clear()
        except: pass
        self.sel_files = [str(self.open_files[i]) for i in self.text_files.curselection()]
        if len(self.sel_files) == 0:
            tkinter.messagebox.showwarning("Inversion error",
                                     "No data selected for inversion \nSelect at least one data file in the left panel", parent=self.master)
        if len(self.sel_files) >= 1:
            try:
                self.Inversion()
                stdout.flush()
            except:
                tkinter.messagebox.showerror("Inversion error", "Error\nMake sure all fields are OK\nMake sure data file is correctly formatted",
                                             parent=self.master)
        return

    def Inversion(self):
        print("\n")
        print("=====================")
        print("Starting inversion...")
        print("=====================")

        print("Model:", self.model.get())
        if self.model.get() == "ColeCole":
            print("Cole-Cole modes:", self.modes_n.get())
        if self.model.get() == "PDecomp":
            print("Polynomial order:", self.poly_n.get())
            if self.c_exp.get() == 1.0:
                decomp_type = "(Debye)"
            elif self.c_exp.get() == 0.5:
                decomp_type = "(Warburg)"
            else:
                decomp_type = "(Cole-Cole)"
            print("Frequency dependence:", self.c_exp.get(), decomp_type)
        print("Units:", self.units.get())
        print("Paths:")
        for i in self.sel_files:
            print(i)
        print("Skipping", self.head.get(), "header lines")
        self.files = [self.sel_files[i].split("/")[-1].split(".")[0] for i in range(len((self.sel_files)))]

        self.mcmc_params = {"adaptive"   : self.adaptive.get(),
                            "verbose"    : self.save_options["Tuning verbose"].get(),
                            "nb_chain"   : self.mcmc_vars["Number of chains"][0].get(),
                            "nb_iter"    : self.mcmc_vars["Total iterations"][0].get(),
                            "nb_burn"    : self.mcmc_vars["Burn-in period"][0].get(),
                            "thin"       : self.mcmc_vars["Thinning factor"][0].get(),
                            "tune_inter" : self.mcmc_vars["Tuning interval"][0].get(),
                            "prop_scale" : self.mcmc_vars["Proposal scale"][0].get(),
                            "cov_inter"  : self.mcmc_vars["Covariance interval"][0].get(),
                            "cov_delay"  : self.mcmc_vars["Covariance delay"][0].get(),
                            }
        # Appel de la fonction d'inversion avec les paramètres sélectionnés
        try:    del(self.all_results)
        except: pass
        self.all_results = {}
        self.draw_drop_down()
        self.var_review.set("Working...")
        for (i, self.f_n) in enumerate(self.files):
            print("=====================")
            self.activity()
            self.var_review.set(self.f_n)
            self.sol = mcmcinv(   self.model.get(), self.sel_files[i], mcmc = self.mcmc_params,
                                headers=self.head.get(), ph_units=self.units.get(),
                                cc_modes=self.modes_n.get(), decomp_poly=self.poly_n.get(),
                                c_exp=self.c_exp.get(), keep_traces=self.save_options["Save traces as txt"].get())

            self.all_results[self.f_n] = {"pm":self.sol["params"],"MDL":self.sol["pymc_model"],"data":self.sol["data"],"fit":self.sol["fit"], "sol":self.sol}
#           Impression ou non des résultats, graphiques, histogrammes
            try:            
                self.update_results()        
            except:
                print("PROBLEM")
            if self.run_options["Print results in console"].get():
                iR.print_resul(self.sol)
            if self.run_options["Save CSV results"].get():
                iR.save_resul(self.sol)
            fig_fit = iR.plot_fit(self.sol, save=self.save_options["Save fit figures"].get(), save_as_png=self.save_options["PNG figures"].get(), draw=self.run_options["Auto draw fit"].get())
            if self.run_options["Auto draw fit"].get():
                self.plot_window(fig_fit, "Inversion results: "+self.f_n)
            if self.model.get() == "PDecomp":
                iR.plot_rtd(self.sol, save=self.save_options["Save fit figures"].get(), save_as_png=self.save_options["PNG figures"].get())
            if self.save_options["Save all hexbins (will make error)"].get():
                for v1, v2 in list(combinations(self.list_of_parameters, 2)):
                    iR.plot_hexbin(self.all_results[self.f_n]["sol"], v1, v2, save=True, save_as_png=self.save_options["PNG figures"].get())
            if self.save_options["Save all bivariate KDE (will make error)"].get():
                for v1, v2 in list(combinations(self.list_of_parameters, 2)):
                    iR.plot_KDE(self.all_results[self.f_n]["sol"], v1, v2, save=True, save_as_png=self.save_options["PNG figures"].get())
            if self.save_options["Save histograms"].get():
                try:
                    iR.plot_histo(self.all_results[self.f_n]["sol"], no_subplots=self.save_options["No subplots"].get(), save=True, save_as_png=self.save_options["PNG figures"].get())
                except:
                    pass
            if self.save_options["Save traces"].get():
                iR.plot_traces(self.all_results[self.f_n]["sol"], no_subplots=self.save_options["No subplots"].get(), save=True, save_as_png=self.save_options["PNG figures"].get())
            if self.save_options["Save summaries"].get():
                iR.plot_summary(self.all_results[self.f_n]["sol"], save=True, save_as_png=self.save_options["PNG figures"].get())
            if self.save_options["Save autocorrelations"].get():
                iR.plot_autocorr(self.all_results[self.f_n]["sol"], save=True, save_as_png=self.save_options["PNG figures"].get())
            if self.save_options["Save deviance"].get():
                iR.plot_deviance(self.all_results[self.f_n]["sol"], save=True, save_as_png=self.save_options["PNG figures"].get())
            if self.save_options["Save loglikelihood"].get():
                iR.plot_logp(self.all_results[self.f_n]["sol"], save=True, save_as_png=self.save_options["PNG figures"].get())

            
            if self.files.index(self.f_n)+1 == len(self.files):
                self.activity(done=True)
                self.diagn_buttons()
                self.write_output_path()
                print("=====================")

#==============================================================================
# Result frame
#==============================================================================
    # Batch progress
    def activity(self, idle=False, done=False):
        try:    self.frame_activ.destroy()
        except: pass
        self.frame_activ = tk.Frame(self.frame_results)
        self.frame_activ.grid(row=0, column=0, sticky=tk.E+tk.W, padx=10, pady=10)
        if idle:
            display = """ \n """
            text_act = tk.Label(self.frame_activ, text=display, anchor=tk.W, justify=tk.LEFT, width=40)
        else:
            display = """Working on:\n%s (#%d/%d)..."""%(self.f_n, self.files.index(self.f_n)+1, len(self.files))
            text_act = tk.Label(self.frame_activ, text=display, anchor=tk.W, justify=tk.LEFT, width=40)
        text_act.grid(row=0, column=0, sticky=tk.W+tk.E+tk.N)
        text_act.update()
        if done:
            text_act.destroy()
            text_done = tk.Label(self.frame_activ, text="""Done\n""", width=40, anchor=tk.W)
            text_done.grid(row=0, column=0, sticky=tk.W+tk.E+tk.N)

    def clear(self, menu=False): # self expl
        self.frame_optimal.destroy()
        self.frame_saved_in.destroy()
        self.frame_diagn_but.destroy()
        if menu:
            self.frame_drop.destroy(), self.frame_activ.destroy()

    def change_file(self):
        self.clear()
        self.update_results()
        self.diagn_buttons()
        self.write_output_path()

    def draw_drop_down(self):
        try:    self.frame_drop.destroy()
        except: pass
        self.frame_drop = tk.Frame(self.frame_results)
        self.frame_drop.grid(row=1, column=0, sticky=tk.E+tk.W, padx=10, pady=3)
        self.frame_drop.columnconfigure(0,weight=1)
        self.var_review = tk.StringVar()
        self.var_review.set(self.files[0])
        optionmenu = tk.OptionMenu(self.frame_drop, self.var_review, *self.files, command = lambda x: self.change_file())
        optionmenu.grid(row=0, column=0, sticky=tk.W+tk.E+tk.S)
        optionmenu.config(bg = "gray97", relief=tk.GROOVE)

    def merge_csv_files(self):
        if len(self.files) > 1:
            iR.merge_results(self.sol, self.files)
            print("=====================")
        else:
            print("Can't merge csv files: Only 1 file inverted in last batch")
            print("=====================")
        stdout.flush()

    def RLD_diagnostic(self):
        MDL = self.all_results[self.var_review.get()]["MDL"]
        if self.model.get() == "Debye": adj = -1
        else: adj = 0
        try:
            keys = [x.__name__ for x in MDL.stochastics]
            for (i, k) in enumerate(keys):
                vect = old_div((MDL.trace(k)[:].size),(len(MDL.trace(k)[:])))
                if vect > 1:
                 keys[i] = [k+"%d"%n for n in range(1+adj,vect+1+adj)]
            keys = list(reversed(sorted(flatten(keys))))
            top_RLD = tk.Toplevel()
            top_RLD.title("Raftery-Lewis diagnostic")
            text_RLD = tk.Text(top_RLD, width=110, font=('courier', fontsize, 'normal'))
            text_RLD.grid(stick=tk.N, padx=(20,0), pady=(10,10))
            q=0.025
            r=0.01
            s=0.95
            for k in keys:
                if k[-1] not in ["%d"%d for d in range(1+adj,8)] or k == "R0":
                    data = MDL.trace(k)[:].ravel()
                else:
                    data = MDL.trace(k[:-1])[:][:,int(k[-1])-1-adj].ravel()
                F="%s:\n"%k
                nmin, kthin, nburn, nprec, kmind = iR.print_diagn(data, q, r, s)
                A="%s iterations required (assuming independence) to achieve %s accuracy with %i percent probability.\n" %(nmin, r, 100 * s)
                B="Thinning factor of %i required to produce a first-order Markov chain.\n" %kthin
                C="%i iterations to be discarded at the beginning of the simulation (burn-in).\n" %nburn
                D="%s subsequent iterations required.\n" %nprec
                E="Thinning factor of %i required to produce an independence chain.\n\n" %kmind
                text_RLD.insert("1.0", F+A+B+C+D+E)
            text_RLD.insert("1.0", "(From PYMC)"+"\n\n")
            text_RLD.insert("1.0", self.var_review.get()+"\n\n")
            button = tk.Button(top_RLD, height=1, width=20, text="Dismiss", command=top_RLD.destroy, bg='gray97', relief=tk.GROOVE)
            button.grid(row=1, column=0, sticky=tk.S, pady=(0,10))

            s = tk.Scrollbar(top_RLD, width=20)
            s.grid(row=0, column=1, sticky=tk.E+tk.N+tk.S, padx=(0,0),pady=(10,10))
            s['command'] = text_RLD.yview
            text_RLD['yscrollcommand'] = s.set
            top_RLD.resizable(width=tk.FALSE, height=tk.FALSE)
        except:
            tkinter.messagebox.showwarning("Diagnostic error",
                                     "Error\nRun inversion first", parent=self.master)

    def popup_bivar(self, bivar_type):
        try: self.top_bivar.destroy()
        except: pass
        self.top_bivar = tk.Toplevel()
        self.top_bivar.title("Bivariate plotting")
        tk.Label(self.top_bivar, text="Select two different parameters: ").grid(row=0, column=0, sticky=tk.W+tk.E+tk.N, pady=(5,5))
        self.biv1, self.biv2 = tk.StringVar(), tk.StringVar()
        self.biv1.set(self.list_of_parameters[0])
        self.biv2.set(self.list_of_parameters[1])
        optionmenu1 = tk.OptionMenu(self.top_bivar, self.biv1, *self.list_of_parameters)
        optionmenu2 = tk.OptionMenu(self.top_bivar, self.biv2, *self.list_of_parameters)
        optionmenu1.grid(row=1, column=0, sticky=tk.W+tk.E+tk.S)
        optionmenu1.config(bg = "gray97", relief=tk.GROOVE)
        optionmenu2.grid(row=2, column=0, sticky=tk.W+tk.E+tk.S)
        optionmenu2.config(bg = "gray97", relief=tk.GROOVE)
        button = tk.Button(self.top_bivar, height=1, width=20, text="OK", command=lambda: self.plot_diagnostic(bivar_type), bg='gray97', relief=tk.GROOVE)
        button.grid(row=3, column=0, sticky=tk.S, pady=(10,10))

    # Diagnostics buttons
    def diagn_buttons(self):
        try:    self.frame_diagn_but.destroy()
        except: pass
        self.frame_diagn_but = tk.Frame(self.frame_results)
        self.frame_diagn_but.grid(row=14, column=0, sticky=tk.E+tk.W+tk.S, padx=10, pady=10)
        self.frame_diagn_but.columnconfigure((0,1,2), weight=1)
        but_cle = tk.Button(self.frame_diagn_but, height=1, width=10, text = "Clear", fg='black', bg='gray97', font=fontz["normal_small"],
                            command = lambda: self.clear(menu=True), relief=tk.GROOVE)
        but_cle.grid(row=9, column=0, columnspan=3, sticky=tk.E+tk.S, pady=(5,0))
        but_mer = tk.Button(self.frame_diagn_but, height=1, text = "Merge csv result files", fg='black', bg='gray97',
                            command = self.merge_csv_files, font=fontz["normal_small"], relief=tk.GROOVE)
        but_mer.grid(row=0, column=0, columnspan=3, sticky=tk.W+tk.E+tk.S, pady=(5,0))
        but_fit = tk.Button(self.frame_diagn_but, height=1, text = "Draw data and fit", fg='black', bg='gray97',
                            command = self.plot_fit_now, font=fontz["normal_small"], relief=tk.GROOVE)
        but_fit.grid(row=1, column=1, sticky=tk.W+tk.E+tk.S, padx=(5,0), pady=(5,0))
        but_dev = tk.Button(self.frame_diagn_but, height=1, text = "Model deviance", fg='black', bg='gray97',
                            command = lambda: self.plot_diagnostic("deviance"), font=fontz["normal_small"], relief=tk.GROOVE)
        but_dev.grid(row=1, column=0, sticky=tk.W+tk.E+tk.S, padx=(0,0), pady=(5,0))
        but_lik = tk.Button(self.frame_diagn_but, height=1, text = "Log-likelihood", fg='black', bg='gray97',
                            command = lambda: self.plot_diagnostic("logp"), font=fontz["normal_small"], relief=tk.GROOVE)
        but_lik.grid(row=1, column=2, sticky=tk.W+tk.E+tk.S, padx=(5,0), pady=(5,0))
        but_sum = tk.Button(self.frame_diagn_but, height=1, text = "Summary and Gelman-Rubin convergence", fg='black', bg='gray97',
                            command = lambda: self.plot_diagnostic("summary"), font=fontz["normal_small"], relief=tk.GROOVE)
        but_sum.grid(row=7, column=0, columnspan=3, sticky=tk.W+tk.E+tk.S, pady=(5,0))
        but_hex = tk.Button(self.frame_diagn_but, height=1, text = "Hex. binning", fg='black', bg='gray97',
                            command = lambda: self.popup_bivar("hexbin"), font=fontz["normal_small"], relief=tk.GROOVE)
        but_hex.grid(row=2, column=0, columnspan=1, sticky=tk.W+tk.E+tk.S, pady=(5,0))
        but_kde = tk.Button(self.frame_diagn_but, height=1, text = "Bivariate KDE", fg='black', bg='gray97',
                            command = lambda: self.popup_bivar("KDE"), font=fontz["normal_small"], relief=tk.GROOVE)
        but_kde.grid(row=2, column=2, columnspan=1, sticky=tk.W+tk.E+tk.S, pady=(5,0), padx=(5,0))
        but_rld = tk.Button(self.frame_diagn_but, height=1, text = "Raftery-Lewis diagnostic", fg='black', bg='gray97',
                            command = self.RLD_diagnostic, font=fontz["normal_small"], relief=tk.GROOVE)
        but_rld.grid(row=8, column=0, columnspan=3, sticky=tk.W+tk.E+tk.S, pady=(5,5))
        but_tra = tk.Button(self.frame_diagn_but, height=1, width=1, text = "Traces", fg='black', bg='gray97',
                            command = lambda: self.plot_diagnostic("traces"), font=fontz["normal_small"], relief=tk.GROOVE)
        but_tra.grid(row=6, column=0, sticky=tk.W+tk.E+tk.S, padx=(0,0), pady=(5,0))
        but_his = tk.Button(self.frame_diagn_but, height=1, width=1, text = "Histograms", fg='black', bg='gray97',
                            command = lambda: self.plot_diagnostic("histo"), font=fontz["normal_small"], relief=tk.GROOVE)
        but_his.grid(row=6, column=1, sticky=tk.W+tk.E+tk.S, padx=(5,0), pady=(5,0))
        but_aut = tk.Button(self.frame_diagn_but, height=1, width=1, text = "Autocorrelation", fg='black', bg='gray97',
                            command = lambda: self.plot_diagnostic("autocorr"), font=fontz["normal_small"], relief=tk.GROOVE)
        but_aut.grid(row=6, column=2, sticky=tk.W+tk.E+tk.S, padx=(5,0), pady=(5,0))
        if self.model.get() == "PDecomp":
            but_rtd = tk.Button(self.frame_diagn_but, height=1, text = "Relaxation time distribution", fg='black', bg='gray97',
                                command = self.plot_rtd_now, font=fontz["normal_small"], relief=tk.GROOVE)
            but_rtd.grid(row=5, column=0, columnspan=3, sticky=tk.W+tk.E+tk.S, pady=(5,0))

    def write_output_path(self):
        try:    self.frame_saved_in.destroy()
        except: pass
        self.frame_saved_in = tk.Frame(self.frame_results)
        self.frame_saved_in.grid(row=8, column=0, sticky=tk.E+tk.W, padx=10, pady=0)
        self.frame_saved_in.columnconfigure(0, weight=1)
        label_done = tk.Label(self.frame_saved_in, text="""Results saved in:""", anchor=tk.W)
        label_done.grid(row=0, column=0, columnspan=1, sticky=tk.W+tk.E)
        text_path = tk.Text(self.frame_saved_in, height=3, width=35, font=fontz["normal_small"])
        text_path.grid(row=1, column=0, columnspan=1, sticky=tk.W+tk.E, padx=0)
        if self.run_options["Save CSV results"].get():         
            text_path.insert("1.0", "%s" %(self.working_path))
        else:
            text_path.insert("1.0", "RESULTS NOT SAVED. CSV SAVE OPTION IS UNCHECKED.")
            
    def update_results(self):
        pm = self.all_results[self.var_review.get()]["pm"]
        model = self.model.get()
        try:    self.frame_optimal.destroy()
        except: pass
        self.frame_optimal = tk.Frame(self.frame_results)
        self.frame_optimal.grid(row=2, column=0, sticky=tk.E+tk.W, padx=10, pady=10)
        self.frame_optimal.columnconfigure(0, weight=1)
        keys = sorted([x for x in list(pm.keys()) if "_std" not in x])
        try:
            keys.remove("m_")
        except:
            pass
        if model == "PDecomp":
            adj = -1
        else:
            adj = 0
        values = flatten([pm[k] for k in sorted(keys)])
        errors = flatten([pm[k+"_std"] for k in sorted(keys)])
        for (i, k) in enumerate(keys):
            if len(pm[k].shape) > 0:
                keys[i] = [keys[i]+"%d"%n for n in range(1+adj,pm[k].shape[0]+1+adj)]
        keys = list(flatten(keys))
        self.list_of_parameters = keys
#        for c, k in enumerate(self.list_of_parameters):
#            if "tau" in k:
#                self.list_of_parameters[c] = "log_"+k
        label_res = tk.Label(self.frame_optimal, text="""Optimal parameters:""", anchor=tk.W)
        label_res.grid(row=0, column=0, sticky=tk.W+tk.E+tk.N)
        text_res = tk.Text(self.frame_optimal, height=len(values), width=40, font=("Courier new", fontsize+res_size, "bold"))
        text_res.grid(row=1, column=0, sticky=tk.W+tk.E+tk.N)
        items = ["{:<13}".format(x+":") for x in keys]
        items2 = [" %.3e " %x for x in values]
        items3 = ["+/- %.0e" %x for x in errors]
        items4 = [" (%.2f%%)" %(abs(100*e/v)) for v,e in zip(values,errors)]
        all_items = [a_+b_+c_+d_ for a_,b_,c_,d_ in zip(items,items2,items3,items4)]
        items = '\n'.join(all_items)
        text_res.insert("1.0", items)

#==============================================================================
# MCMC parameters
#==============================================================================
    def mcmc_parameters(self):
        # McMC parameters
        self.adaptive = tk.BooleanVar()
        self.adaptive_text = tk.StringVar()
        def yesno():
            if self.adaptive.get():
                self.adaptive_text.set("AM")
                for i, (k, v) in enumerate(self.mcmc_vars.items()):
                    v[0].set(self.root_ini[k])
                    if k in ["Tuning interval", "Proposal scale"]:
                        tk.Entry(self.frame_mcmc ,textvariable=v[0],fg="grey", width=12).grid(row=i+1, column=1, sticky=tk.E,padx=(0,10))
                        tk.Label(self.frame_mcmc, text=k, justify = tk.LEFT, fg="grey").grid(row=i+1, column=0, sticky=tk.W, padx=(10,0))
                    else:
                        tk.Entry(self.frame_mcmc ,textvariable=v[0], width=12).grid(row=i+1, column=1, sticky=tk.E,padx=(0,10))
                        tk.Label(self.frame_mcmc, text=k, justify = tk.LEFT).grid(row=i+1, column=0, sticky=tk.W, padx=(10,0))
            else:
                self.adaptive_text.set("MH")
                for i, (k, v) in enumerate(self.mcmc_vars.items()):
                    v[0].set(self.root_ini[k])
                    if k in ["Covariance interval", "Covariance delay"]:
                        tk.Entry(self.frame_mcmc ,textvariable=v[0],fg="grey", width=12).grid(row=i+1, column=1, sticky=tk.E,padx=(0,10))
                        tk.Label(self.frame_mcmc, text=k, justify = tk.LEFT, fg="grey").grid(row=i+1, column=0, sticky=tk.W, padx=(10,0))
                    else:
                        tk.Entry(self.frame_mcmc ,textvariable=v[0], width=12).grid(row=i+1, column=1, sticky=tk.E,padx=(0,10))
                        tk.Label(self.frame_mcmc, text=k, justify = tk.LEFT).grid(row=i+1, column=0, sticky=tk.W, padx=(10,0))

        tk.Label(self.frame_mcmc, text="Use Adaptive Metropolis:", justify = tk.LEFT).grid(row=0, column=0, sticky=tk.W, padx=(10,0))
        tk.Checkbutton(self.frame_mcmc , variable=self.adaptive, textvariable=self.adaptive_text, width=12, command=yesno).grid(row=0, column=1, sticky=tk.E,padx=(0,10))
        self.adaptive.set(self.root_ini["Adaptive Metropolis"])
        yesno()

#==============================================================================
# SIP model
#==============================================================================
    def model_choice(self):
        # Available models
        models = [("Pelton \nCole-Cole","ColeCole"),
                  ("Dias \nmodel","Dias"),
                  ("Debye / Warburg \ndecomposition","PDecomp"),]
        self.modes_n, self.poly_n, self.c_exp = tk.IntVar(), tk.IntVar(), tk.DoubleVar()
        self.modes_n.set(self.root_ini["Nb modes"]), self.poly_n.set(self.root_ini["Polyn order"]), self.c_exp.set(self.root_ini["Freq dep"])

        ### Model choice
        self.model = tk.StringVar()
        self.model.set(self.root_ini["Spectral IP model"])  # set last used values
        for i, (txt, val) in enumerate(models):
            tk.Radiobutton(self.frame_model, text=txt, justify=tk.LEFT, variable = self.model, command = self.draw_rtd_check, value=val).grid(row=i, column=0, sticky=tk.W+tk.S, padx=(10,0), pady=pad_radio+2)
        self.draw_rtd_check()

    def draw_rtd_check(self):
        try:
            self.mod_opt_frame.destroy()
        except:
            pass
        self.mod_opt_frame = tk.Frame(self.frame_model)
        self.mod_opt_frame.grid(row=0, column=1, rowspan=4)
        self.mod_opt_frame.grid_rowconfigure(4, weight=1)
        if self.model.get() == "PDecomp":
            poly_lab = tk.Label(self.mod_opt_frame, text="""Polyn order""", justify = tk.LEFT, font=fontz["normal_small"])
            poly_lab.grid(row=0, column=1, rowspan=1, sticky=tk.W+tk.N, pady=(0,0), padx=(0,10))
            poly_scale = tk.Scale(self.mod_opt_frame, variable=self.poly_n, width=10, length=70, from_=3, to=6, font=fontz["normal_small"], orient=tk.HORIZONTAL)
            poly_scale.grid(row=1, column=1, rowspan=1, sticky=tk.E+tk.N, padx=(0,10), pady=(0,0))
            exp_lab = tk.Label(self.mod_opt_frame, text="""c exponent""", justify = tk.LEFT, font=fontz["normal_small"])
            exp_lab.grid(row=2, column=1, rowspan=1, sticky=tk.W+tk.N, pady=(0,0), padx=(0,10))
            exp_scale = tk.Scale(self.mod_opt_frame, variable=self.c_exp, width=10, length=70, from_=0.1, to=1.0, resolution=0.05, font=fontz["normal_small"], orient=tk.HORIZONTAL)
            exp_scale.grid(row=3, column=1, rowspan=1, sticky=tk.E+tk.N, padx=(0,10), pady=(0,0))
        if self.model.get() == "ColeCole":
            modes_lab = tk.Label(self.mod_opt_frame, text="""Nb modes""", justify = tk.LEFT, font=fontz["normal_small"])
            modes_lab.grid(row=0, column=1, rowspan=1, sticky=tk.W+tk.N, pady=(0,0), padx=(0,10))
            modes_scale = tk.Scale(self.mod_opt_frame, variable=self.modes_n, width=10, length=70, from_=1, to=3, font=fontz["normal_small"], orient=tk.HORIZONTAL)
            modes_scale.grid(row=1, column=1, rowspan=1, sticky=tk.E+tk.N, padx=(0,10), pady=(0,0))

#==============================================================================
# Run Exit Options
#==============================================================================
    def run_exit(self):
        for i, (k, v) in enumerate(self.run_options.items()):
            tk.Checkbutton(self.frame_ruex, text=k, variable=v).grid(row=i, column=0, padx=(10,0), sticky=tk.W+tk.N+tk.S)
            self.run_options[k].set(self.root_ini[k]) # set last used values
        tk.Button(self.frame_ruex, width=14, text = "Options", fg='black', bg='gray97', font=fontz["bold"], relief=tk.GROOVE,
                            command = self.make_options).grid(row=i+1, column=0, rowspan=1, sticky=tk.N+tk.E+tk.W+tk.S, padx=(10,5), pady=(0,10))
        tk.Button(self.frame_ruex, width=14, height=1, text = "RUN", fg='blue', bg='gray97', font=fontz["bold"], relief=tk.GROOVE,
                            command = self.run_inversion).grid(row=0, column=1, rowspan=2, sticky=tk.N+tk.E+tk.S, padx=(5,10), pady=(5,0), ipadx=0, ipady=0)
        tk.Button(self.frame_ruex, width=14, height=1, text = "EXIT", fg='red', bg='gray97', font=fontz["bold"], relief=tk.GROOVE,
                            command = self.master.destroy).grid(row=2, column=1, rowspan=2, sticky=tk.S+tk.E+tk.N, padx=(5,10), pady=(0,10))

    def make_options(self, from_root=False):
        try: self.top_checkbox.destroy()
        except: pass

        # Select all or none functions
        def select_all(): # self expl
            for v in list(self.save_options.values()):
                v.set(True)
        def select_none(): # self expl
            for v in list(self.save_options.values()):
                v.set(False)
        if not from_root:
            self.top_checkbox = tk.Toplevel(self.master)
            self.top_checkbox.title("Save options")
            frame_checkbox = tk.LabelFrame(self.top_checkbox, text="Check items to save", width=200, height=4, font=fontz["bold"])
            frame_checkbox.grid(row = 0, column=0,sticky=tk.S+tk.W+tk.E+tk.N, padx=10, pady=10)
            tk.Button(frame_checkbox, height=1, width=15, text="Check all", command=select_all, bg='gray97', font=fontz["normal_small"], relief=tk.GROOVE).grid(row=0, column=0, sticky=tk.W, pady=5, padx=10)
            tk.Button(frame_checkbox, height=1, width=15, text="Check none", command=select_none, bg='gray97', font=fontz["normal_small"], relief=tk.GROOVE).grid(row=0, column=1, sticky=tk.E, pady=5, padx=10)
            button = tk.Button(frame_checkbox, height=2, width=20, text="OK", bg='gray97', font=fontz["bold"], command=self.top_checkbox.destroy, relief=tk.GROOVE)
            button.grid(row=len(self.save_options)+2, column=0, columnspan=2, sticky=tk.W+tk.E, pady=(10,10), padx=(20,20))
        for i, (k, v) in enumerate(sorted(self.save_options.items())):
            if from_root:
                self.save_options[k].set(self.root_ini[k])
            if not from_root:
                tk.Checkbutton(frame_checkbox, text=k, variable=v).grid(row=i+1, column=0, sticky=tk.W+tk.N+tk.S,padx=(10,10),pady=(5,0))

#==============================================================================
# Plotting
#==============================================================================
    def set_plot_par(self): # Setting up plotting parameters
        try:
            print("\nLoading plot parameters")
            rcParams.update(iR.plot_par())
            print("Plot parameters successfully loaded")
        except:
            print("Plot parameters not found, using default values")
        stdout.flush()

    def preview_data(self):
        try:
            for i in self.text_files.curselection():
                sel = str(self.open_files[i])
                fn = sel.split("/")[-1].split(".")[0]
                fig_data = iR.plot_data(sel, self.head.get(), self.units.get())
                self.plot_window(fig_data, "Data preview: "+fn)
        except:
            tkinter.messagebox.showwarning("Preview error",
                                     "Can't draw data\nImport and select at least one data file first", parent=self.master)

    def plot_diagnostic(self, which):
        f_n = self.var_review.get()
        sol = self.all_results[f_n]["sol"]
        try:
            if which == "traces":
                trace_plot = iR.plot_traces(sol, save=False)
                self.plot_window(trace_plot, "Parameter traces: "+f_n)
            if which == "histo":
                histo_plot = iR.plot_histo(sol, save=False)
                self.plot_window(histo_plot, "Parameter histograms: "+f_n)
            if which == "autocorr":
                autocorr_plot = iR.plot_autocorr(sol, save=False)
                self.plot_window(autocorr_plot, "Parameter autocorrelation: "+f_n)
            if which == "geweke":
                geweke_plot = iR.plot_scores(sol, save=False)
                self.plot_window(geweke_plot, "Geweke scores: "+f_n)
            if which == "summary":
                summa_plot = iR.plot_summary(sol, save=False)
                self.plot_window(summa_plot, "Parameter summary: "+f_n)
            if which == "deviance":
                devi_plot = iR.plot_deviance(sol, save=False)
                self.plot_window(devi_plot, "Model deviance: "+f_n)
            if which == "logp":
                logp_plot = iR.plot_logp(sol, save=False)
                self.plot_window(logp_plot, "Log-likelihood: "+f_n)
            if which == "hexbin":
                try: self.top_bivar.destroy()
                except: pass
                hex_plot = iR.plot_hexbin(sol, self.biv1.get(), self.biv2.get(), save=False)
                self.plot_window(hex_plot, "Hexagonal binning: "+f_n)
            if which == "KDE":
                try: self.top_bivar.destroy()
                except: pass
                kde_plot = iR.plot_KDE(sol, self.biv1.get(), self.biv2.get(), save=False)
                self.plot_window(kde_plot, "Bivariate KDE: "+f_n)
            stdout.flush()
        except:
            tkinter.messagebox.showwarning("Error analyzing results", "Error\nProblem with inversion results\nTry adding iterations",
                                     parent=self.master)

    def plot_window(self, fig, name=""):
        top_fig = tk.Toplevel()
        top_fig.lift()
        top_fig.title(name)
        top_fig.rowconfigure(1, weight=1)
        def _quit():
            top_fig.destroy()
        canvas = FigureCanvasTkAgg(fig, master=top_fig)
        canvas.show()
        canvas.get_tk_widget().grid(row=1, column=0, columnspan = 3, pady=(15,15), padx=(25,25), sticky=tk.N+tk.S+tk.E+tk.W)
        button = tk.Button(top_fig, height=1, width=20, text="Dismiss", bg='gray97', command=_quit, relief=tk.GROOVE)
        button.grid(row=2, column=0, columnspan=1, sticky=tk.W, pady=(10,10), padx=(20,20))
        toolbar_frame = tk.Frame(top_fig)
        toolbar_frame.grid(row=0,column=0,columnspan=2, sticky=tk.W+tk.E)
        NavigationToolbar2TkAgg( canvas, toolbar_frame )
        top_fig.resizable(width=tk.FALSE, height=tk.FALSE)

    def plot_fit_now(self):
        f_n = self.var_review.get()
        sol = self.all_results[f_n]["sol"]
        fig_fit = iR.plot_fit(sol, save=False, draw=True)
        self.plot_window(fig_fit, "Inversion results: "+f_n)

    def plot_rtd_now(self):
        f_n = self.var_review.get()
        fig_debye = iR.plot_rtd(self.all_results[f_n]["sol"], save=False, draw=True)
        self.plot_window(fig_debye, "Debye RTD: "+f_n)

#==============================================================================
# Importing data
#==============================================================================

    def make_browse_button(self):
        self.but_browse = tk.Button(self.frame_import, height=1, text = "Open file browser", bg='gray97',
                            command = self.get_file_list, relief=tk.GROOVE)
        self.but_browse.grid(row=0, column=0, columnspan=2, pady=(10,10), padx=(10,10), sticky=tk.E+tk.W)

    def headers_input(self):
        self.head = tk.IntVar()
        self.head.set(self.root_ini["Nb header lines"])  # set last used value
        tk.Label(self.frame_import, text="""Nb header lines:""", justify = tk.LEFT).grid(row=1, column=0, columnspan=1,sticky=tk.W, padx=(10,0))
        tk.Entry(self.frame_import, textvariable=self.head, width=12).grid(row=1, column=1, columnspan=1,sticky=tk.E+tk.W,padx=(10,10))

    def phase_units(self):
        # Phase units
        unites = ["mrad", "rad" ,"deg"]
        self.units = tk.StringVar()
        self.units.set(self.root_ini["Phase units"])  # set last used value
        tk.Label(self.frame_import, text="""Phase units:""", justify = tk.LEFT).grid(row=2, column=0, sticky=tk.W, padx=(10,0))
        for i, u in enumerate(unites):
            tk.Radiobutton(self.frame_import, text=u, variable=self.units, command=None, value=u).grid(row=i+2, column=1, sticky=tk.W, padx=(10,10), pady=pad_radio)

    def get_file_list(self):
        files = tkinter.filedialog.askopenfilenames(parent=self.master,title='Choose a file')
        self.open_files = sorted(list(files))
        self.draw_file_list()

    def draw_file_list(self):

        self.frame_path = tk.Frame(self.frame_list)
        self.frame_path.grid(row=0, column=0, padx=20, pady=10, sticky=tk.E+tk.W)
        self.frame_path.columnconfigure(2, weight=1)

        self.frame_select = tk.Frame(self.frame_list)
        self.frame_select.grid(row=1, column=0, padx=(20,0), pady=10, sticky=tk.E+tk.W+tk.S+tk.N)
        self.frame_select.columnconfigure(0, weight=1)
        self.frame_select.rowconfigure(1, weight=1)

        text_loc = tk.Text(self.frame_path, width=40, height=3, font=fontz["normal_small"])
        text_loc.grid(row=0, column=0, rowspan=1, sticky=tk.N+tk.W+tk.E)
        s = tk.Scrollbar(self.frame_select, width=20)
        s.grid(row=1, column=2, sticky=tk.E+tk.N+tk.S, pady=(0,10))
        self.text_files = tk.Listbox(self.frame_select, selectmode='extended', font=fontz["italic_small"])
        self.text_files.grid(row=1, column=0, columnspan=2, sticky=tk.S+tk.W+tk.E+tk.N, pady=(0,0))
        s['command'] = self.text_files.yview
        self.text_files['yscrollcommand'] = s.set

        tk.Button(self.frame_select, height=1, text="Preview selected data files", command=self.preview_data, bg='gray97', font=fontz["normal_small"], relief=tk.GROOVE).grid(row=2, column=0, columnspan=2, sticky=tk.W+tk.E)
        # Select all or none functions
        def select_all(): # self expl
            self.text_files.select_set(0, tk.END)
        def select_none(): # self expl
            self.text_files.select_clear(0, tk.END)
        tk.Button(self.frame_select, height=1, width=15, text="Select all", command=select_all, bg='gray97', font=fontz["normal_small"], relief=tk.GROOVE).grid(row=0, column=0, sticky=tk.W+tk.N, pady=(0,0))
        tk.Button(self.frame_select, height=1, width=15, text="Select none", command=select_none, bg='gray97', font=fontz["normal_small"], relief=tk.GROOVE).grid(row=0, column=1, sticky=tk.E+tk.N, pady=(0,0))

        # Fill list box
        if len(self.open_files) >= 1: # Only draw this listbox if files were selected in browser
            for (i, o) in enumerate(self.open_files):
                fil = o.split("/")[-1] # Split full path to keep only name and extension
                self.text_files.insert(i, fil) # Insert items 1 by 1 in loop
            locs = "/".join(self.open_files[0].split("/")[:-1])+"/" # Path to imported files in string format
            text_loc.insert("1.0", locs) # Insert path to empty text box
            self.text_files.select_set(0) # Select first item in list after populating
        else:
            text_loc.insert("1.0", "None imported") # Empty list box if no files selected

#==============================================================================
# Save and load GUI state
#==============================================================================

    def load(self):
        print("\nLoading root_ini from:\n", self.working_path)
        try:
            with open(self.working_path+'root_ini') as f:
                self.root_ini = jload(f)
            print("root_ini successfully loaded")
        except:
            print("root_ini not found, using default values")
            self.root_ini = self.use_default_root_ini()
        stdout.flush()

    def save(self):
        root_save = {
                        "Spectral IP model"     : self.model.get(),
                        "Adaptive Metropolis"   : self.adaptive.get(),
                        "Nb header lines"       : self.head.get(),
                        "Phase units"           : self.units.get(),
                        "Imported files"        : self.open_files,
                        "Polyn order"           : self.poly_n.get(),
                        "Freq dep"              : self.c_exp.get(),
                        "Nb modes"              : self.modes_n.get(),
                    }

        for k, v in list(self.mcmc_vars.items()):
            root_save[k] = v[0].get()
        for k, v in list(self.save_options.items()):
            root_save[k] = v.get()
        for k, v in list(self.run_options.items()):
            root_save[k] = v.get()

        print("\nSaving root_ini")
        with open(self.working_path+'root_ini', 'w') as f:
            jdump(root_save, f)
        print("root_ini successfully saved in:\n", self.working_path)
        print("\n=========END=========")

#==============================================================================
# Build menubar File, Help
#==============================================================================
    def build_helpmenu(self):
        menubar = tk.Menu(self.master)
        filemenu = tk.Menu(menubar, tearoff=0)
        helpmenu = tk.Menu(menubar, tearoff=0)
        filemenu.add_command(label="Open", command=self.get_file_list)
        filemenu.add_separator()
        filemenu.add_command(label="Exit", command=self.master.destroy)
        menubar.add_cascade(label="File", menu=filemenu)
        helpmenu.add_command(label="Data file template", command=lambda: TextMessage().popup("Data file template", TextMessage.data_template, size=fontsize))
        helpmenu.add_command(label="MCMC parameters", command=lambda: TextMessage().popup("Markov-chain Monte Carlo parameters", TextMessage.mcmc_info, size=fontsize))
        helpmenu.add_command(label="References", command=lambda: TextMessage().popup("References", TextMessage.references, size=fontsize))
        helpmenu.add_separator()
        helpmenu.add_command(label="License", command=lambda: TextMessage().popup("License information", TextMessage.license_info, size=fontsize))
        helpmenu.add_command(label="About", command=lambda : TextMessage().about(fontz))
        menubar.add_cascade(label="Help", menu=helpmenu)
        self.master.config(menu=menubar)



#==============================================================================
# Class for popup messages
#==============================================================================
class TextMessage(object):

    data_template = """DATA FILE TEMPLATE

Save data in .csv, .txt, .dat, ... extension file
Comma separation between columns is mandatory
Column order is very important
Phase units may be milliradians, radians or degrees
Units are specified in main window
Amplitude units may be Ohm-m or Ohm
A number of header lines may be skipped in the main window
In this example Nb header lines = 1
To skip high-frequencies, increase Nb header lines
Scientific or standard notation is OK
Data must be formatted using the following template:

===============================================================================

Freq (Hz), Res (Ohm-m),  Phase (deg), dRes (Ohm-m), dPhase (deg)
6.000e+03, 1.17152e+05, -2.36226e+02, 1.171527e+01, 9.948376e-02
3.000e+03, 1.22177e+05, -1.46221e+02, 1.392825e+01, 1.134464e-01
1.500e+03, 1.25553e+05, -9.51099e+01, 2.762214e+01, 2.199114e-01
........., ..........., ............, ............, ............
........., ..........., ............, ............, ............
........., ..........., ............, ............, ............
4.575e-02, 1.66153e+05, -1.21143e+01, 1.947314e+02, 1.171115e+00
2.288e-02, 1.67988e+05, -9.36718e+00, 3.306003e+02, 1.9669860+00
1.144e-02, 1.70107e+05, -7.25533e+00, 5.630541e+02, 3.310889e+00

===============================================================================
"""

    mcmc_info = """===================================================================================
Meaning of MCMC parameters:

|---------------------------# Iterations------------------------------------| = 100
|........Burn-in period (discarded iterations)........|---------------------| =  79
         Remaining iterations:                        |---------------------| =  21
         Thinning factor (Discard more chains):       |-..-..-..-..-..-..-..| =   3
         Iterations kept for histogram:                             |-------| =   7

         Tuning interval = 1000: the proposal scale is tuned every 1000 iteration
         Proposal scale: the proposal distribution std = scale*parameter_value
         Adding more chains may help sample the posterior distribution
         More iterations will lead to better fit up until a certain point
         Burn-in to discard the random hypotheses before convergence
         Burn-in period is used for tuning
         Thinning factor has no advantage for convergence or quality of fit

         Burn-in for 90% of iterations
         Use thinning factor to keep ~1000-10000 samples in histograms
         (easier on memory)

===================================================================================
Recommended for data files with ~20 frequencies:

Cole-Cole model:    Good convergence for 1 or 2 modes
                    Slow convergence for 3 modes
                    10 000 - 100 000 iterations for 2 modes
                    May take > 500 000 for 3 modes

Debye model:        Good convergence for 3rd order polynomial
                    100 000 iterations usually converges for 3rd order
                    > 100 000 iterations for 4th order polynomial
                    Slow convergence, sometimes no convergence for 5th order

Shin model:         10 000 - 100 000 iterations usually converges

Dias model:         Fast convergence, weak fits
                    10 000 iterations usually converges

===================================================================================
"""

    references = """Chen, Jinsong, Andreas Kemna, and Susan S. Hubbard. 2008. “A Comparison between
    Gauss-Newton and Markov-Chain Monte Carlo–based Methods for Inverting
    Spectral Induced-Polarization Data for Cole-Cole Parameters.” Geophysics
    73 (6): F247–59. doi:10.1190/1.2976115.
Dias, Carlos A. 2000. “Developments in a Model to Describe Low-Frequency
    Electrical Polarization of Rocks.” Geophysics 65 (2): 437–51.
    doi:10.1190/1.1444738.
Gamerman, Dani, and Hedibert F. Lopes. 2006. Markov Chain Monte Carlo:
    Stochastic Simulation for Bayesian Inference, Second Edition. CRC Press.
Ghorbani, A., C. Camerlynck, N. Florsch, P. Cosenza, and A. Revil. 2007.
    “Bayesian Inference of the Cole–Cole Parameters from Time- and Frequency-
    Domain Induced Polarization.” Geophysical Prospecting 55 (4): 589–605.
    doi:10.1111/j.1365-2478.2007.00627.x.
Hoff, Peter D. 2009. A First Course in Bayesian Statistical Methods. Springer
    Science & Business Media.
Keery, John, Andrew Binley, Ahmed Elshenawy, and Jeremy Clifford. 2012.
    “Markov-Chain Monte Carlo Estimation of Distributed Debye Relaxations in
    Spectral Induced Polarization.” Geophysics 77 (2): E159–70.
    doi:10.1190/geo2011-0244.1.
Nordsiek, Sven, and Andreas Weller. 2008. “A New Approach to Fitting Induced-
    Polarization Spectra.” Geophysics 73 (6): F235–45. doi:10.1190/1.2987412.
Patil, A., D. Huard and C.J. Fonnesbeck. 2010. PyMC: Bayesian Stochastic
    Modelling in Python. Journal of Statistical Software, 35(4), pp. 1-81
Pelton, W. H., W. R. Sill, and B. D. Smith. 1983. Interpretation of Complex
    Resistivity and Dielectric Data — Part 1. Vol 29. Geophysical Transactions.
Pelton, W. H., S. H. Ward, P. G. Hallof, W. R. Sill, and P. H. Nelson. 1978.
    “Mineral Discrimination and Removal of Inductive Coupling with
    Multifrequency IP.” Geophysics 43 (3): 588–609. doi:10.1190/1.1440839."""

    license_info = """The MIT License (MIT)

Copyright (c) 2016 Charles L. Bérubé

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.

https://opensource.org/licenses/MIT
https://github.com/clberube/bisip
"""

    def popup(self, title, message, size=10):
        top = tk.Toplevel()
        top.title(title)
        about_message = (message)
        top.lift()
        msg = tk.Text(top, width=90, font=('courier', size, 'normal'))
        msg.grid(stick=tk.N, padx=(10,10), pady=(10,10))
        msg.insert("1.0", about_message)
        button = tk.Button(top, height=1, width=20, text="Dismiss", command=top.destroy, bg='gray97', relief=tk.GROOVE)
        button.grid(sticky=tk.S, pady=(0,10))
        s = tk.Scrollbar(top, width=20)
        s.grid(row=0, column=0, sticky=tk.E+tk.N+tk.S, padx=(0,10),pady=(10,10))
        s['command'] = msg.yview
        msg['yscrollcommand'] = s.set
        top.resizable(width=tk.FALSE, height=tk.FALSE)

    def about(self, fontz):
        top = tk.Toplevel()
        top.title("About")
        top.lift()
        tk.Label(top,
                  text="""Bayesian inversion of spectral induced polarization data""",
                  justify = tk.CENTER, font=fontz["bold"]).grid(row=0, column=0,columnspan=2,pady=(10,10), padx=(10,10))
        tk.Label(top,
                  text="""2015-2016
    École Polytechnique de Montréal
    Groupe de recherche en géophysique appliquée
    """,
                  justify = tk.CENTER).grid(row=1, column=0,columnspan=2, pady=(10,10), padx=(10,10))
        tk.Label(top,
                  text="""Contact:""",
                  justify = tk.CENTER).grid(row=2, column=0,columnspan=1, pady=(10,10), padx=(10,10))
        about_message = """cberube@ageophysics.com"""
        msg = tk.Text(top, height=1, width=27)
        msg.grid(row=2,column=1,padx=(10,10), pady=(10,10))
        msg.insert("1.0", about_message)
        button = tk.Button(top, height=1, width=20, text="Dismiss", command=top.destroy, bg='gray97', relief=tk.GROOVE)
        button.grid(row=3,column=0,columnspan=2,sticky=tk.S, pady=(0,10))
        top.resizable(width=tk.FALSE, height=tk.FALSE)

#==============================================================================
# Main function
#==============================================================================
def launch():
    root = tk.Tk()
    root.wm_title("Bayesian inversion of SIP data")
    root.option_add("*Font", window_font)
    #==============================================================================
    # For MacOS, bring the window to front
    # Without these lines the application will start in background
    root.lift()
    if "Darwin" in system():
        root.call('wm', 'attributes', '.', '-topmost', True)
        root.after_idle(root.call, 'wm', 'attributes', '.', '-topmost', False)
    app = MainApplication(root, fontz)
    root.mainloop()
    app.save()
if __name__ == '__launch__':
    launch()


# -*- coding: utf-8 -*-
"""
Created on Fri Sep 29 16:18:50 2017

@author: Charles
"""

from setuptools import setup, find_packages

from distutils.core import setup
from distutils.extension import Extension
import numpy

try:
    from Cython.Distutils import build_ext
except ImportError:
    use_cython = False
else:
    use_cython = True

cmdclass = { }
ext_modules = [ ]

if use_cython:
    ext_modules += [
        Extension("bisip.cython_funcs", [ "bisip/cython_funcs.pyx" ]),
    ]
    cmdclass.update({ 'build_ext': build_ext })
    print("up")
else:
    ext_modules += [
        Extension("bisip.cython_funcs", [ "bisip/cython_funcs.c" ],
                  include_dirs=[numpy.get_include()]),
    ]
    print("down")

setup(
  name = 'bisip',
  packages=['bisip',], # this must be the same as the name above
  py_models=['models','invResults','GUI'],
  version = '0.0.7',
  license = 'MIT',
  install_requires=['pymc'],
  description = 'Bayesian inversion of SIP data',
  long_description = 'README.md',
  author = 'Charles L. Berube',
  author_email = 'cberube@ageophysics.com',
  url = 'https://github.com/clberube/BISIP', # use the URL to the github repo
#  download_url = 'https://github.com/clberube/BISIP/archive/0.0.1.tar.gz', # I'll explain this in a second
  keywords = ['testing', 'logging', 'example'], # arbitrary keywords
  classifiers = [],
  cmdclass = cmdclass,
  ext_modules = ext_modules,
  include_dirs=[numpy.get_include()],
)
